import logging
import pickle
import warnings
from typing import Any, Dict, List, Optional, Type, Union

import mlflow
import mlflow.data

from hamilton import graph_types
from hamilton.lifecycle import GraphConstructionHook, GraphExecutionHook, NodeExecutionHook

# silence odd ongoing MLFlow issue that spams warnings
# GitHub Issue https://github.com/mlflow/mlflow/issues/8605
warnings.filterwarnings("ignore", category=UserWarning)


FIGURE_TYPES = []
try:
    import matplotlib.figure

    FIGURE_TYPES.append(matplotlib.figure.Figure)
except ImportError:
    pass

try:
    import plotly.graph_objects

    FIGURE_TYPES.append(plotly.graph_objects.Figure)
except ImportError:
    pass


logger = logging.getLogger(__name__)


def get_path_from_metadata(metadata: dict) -> Union[str, None]:
    """Retrieve the `path` attribute from DataSaver output metadata"""
    path = None
    if "path" in metadata:
        path = metadata["path"]
    elif "file_metadata" in metadata:
        path = metadata["file_metadata"]["path"]

    return path


# NOTE `mlflow.client.MLFlowClient` is preferred to top-level `mlflow.` methods in MLFlowTracker
# because the latter relies on hard-to-debug global variables. Yet, we set an `active_run` by using
# `mlflow.start_run()` in pre_graph_execution to ensure the user-specified MLFlow code
# and MLFlow materializers log metrics and models to the same run as the MLFlowTracker
class MLFlowTracker(
    NodeExecutionHook,
    GraphExecutionHook,
    GraphConstructionHook,
):
    """Driver adapter logging Hamilton execution results to an MLFlow server."""

    def __init__(
        self,
        tracking_uri: Optional[str] = None,
        registry_uri: Optional[str] = None,
        artifact_location: Optional[str] = None,
        experiment_name: str = "Hamilton",
        experiment_tags: Optional[dict] = None,
        experiment_description: Optional[str] = None,
        run_id: Optional[str] = None,
        run_name: Optional[str] = None,
        run_tags: Optional[dict] = None,
        run_description: Optional[str] = None,
        log_system_metrics: bool = False,
    ):
        """Configure the MLFlow client and experiment for the lifetime of the tracker

        :param tracking_uri: Destination of the logged artifacts and metadata. It can be a filesystem, database, or server. [reference](https://mlflow.org/docs/latest/getting-started/tracking-server-overview/index.html)
        :param registry_uri: Destination of the registered models. By default it's the same as the tracking destination, but they can be different. [reference](https://mlflow.org/docs/latest/getting-started/registering-first-model/index.html)
        :param artifact_location: Root path on tracking server where experiment is stored
        :param experiment_name: MLFlow experiment name used to group runs.
        :param experiment_tags: Tags to query experiments programmatically (not displayed).
        :param experiment_description: Description of the experiment displayed
        :param run_id: Run id to log to an existing run (every execution logs to the same run)
        :param run_name: Run name displayed and used to query runs. You can have multiple runs with the same name but different run ids.
        :param run_tags: Tags to query runs and appears as columns in the UI for filtering and grouping. It automatically includes serializable inputs and Driver config.
        :param run_description: Description of the run displayed
        :param log_system_metrics: Log system metrics to display (requires additonal dependencies)
        """
        self.client = mlflow.client.MlflowClient(tracking_uri, registry_uri)

        # experiment setup
        experiment_tags = experiment_tags if experiment_tags else {}
        if experiment_description:
            # mlflow.note.content is the description field
            experiment_tags["mlflow.note.content"] = experiment_description

        # TODO link HamiltonTracker project and MLFlowTracker experiment
        experiment = self.client.get_experiment_by_name(experiment_name)
        if experiment:
            experiment_id = experiment.experiment_id
            # update tags and description of an existing experiment
            if experiment_tags:
                for k, v in experiment_tags.items():
                    self.client.set_experiment_tag(experiment_id, key=k, value=v)
        # create an experiment
        else:
            experiment_id = self.client.create_experiment(
                name=experiment_name,
                artifact_location=artifact_location,
                tags=experiment_tags,
            )
        self.experiment_id = experiment_id

        # run setup
        # TODO link HamiltonTracker and MLFlowTracker run ids
        self.mlflow_run_id = run_id
        self.run_name = run_name
        self.run_tags = run_tags if run_tags else {}
        if run_description:
            # mlflow.note.content is the description field
            self.run_tags["mlflow.note.content"] = run_description

        self.log_system_metrics = log_system_metrics

    def run_after_graph_construction(self, *, config: dict[str, Any], **kwargs):
        """Store the Driver config before creating the graph"""
        self.config = config

    def run_before_graph_execution(
        self,
        *,
        run_id: str,
        final_vars: List[str],
        inputs: Dict[str, Any],
        graph: graph_types.HamiltonGraph,
        **kwargs,
    ):
        """Create and start MLFlow run. Log graph version, run_id, inputs, overrides"""
        # add Hamilton metadata to run tags
        run_tags = self.run_tags
        run_tags["hamilton_run_id"] = run_id  # the Hamilton run_id
        run_tags["code_version"] = graph.version

        # create Hamilton run
        self.run = self.client.create_run(
            experiment_id=self.experiment_id,
            tags=run_tags,
            run_name=self.run_name,
        )
        self.run_id = self.run.info.run_id
        # start run to set `active_run` and allow user-defined callbacks and materializers
        # to log to the same run as the HamiltonTracker
        mlflow.start_run(
            run_id=self.run_id,
            experiment_id=self.experiment_id,
            tags=run_tags,
            log_system_metrics=self.log_system_metrics,
        )

        # log config to artifacts
        self.client.log_dict(self.run_id, self.config, "config.json")

        # log HamiltonGraph to reproduce the run
        self.graph = graph
        graph_as_json = {n.name: n.as_dict() for n in graph.nodes}
        self.client.log_dict(self.run_id, graph_as_json, "hamilton_graph.json")

        # log config and inputs as `param` which creates columns in the UI to filter runs
        # `log_param()` accepts `value: Any` and will stringify complex objects
        for value_sets in [self.config, inputs]:
            for node_name, value in value_sets.items():
                self.client.log_param(self.run_id, key=node_name, value=value)

        self.final_vars = final_vars

    # TODO log DataLoaders as MLFlow datasets
    def run_after_node_execution(
        self,
        *,
        node_name: str,
        node_return_type: Type,
        node_tags: dict,
        node_kwargs: dict,
        result: Any,
        **kwargs,
    ):
        """Log materializers and final vars as artifacts"""
        # log DataSavers as artifacts
        # TODO refactor if/else as `handle_materializers()`
        if node_tags.get("hamilton.data_saver") is True:
            # don't log mlflow materializers as artifact since they already create models
            # instead, use the Materializer metadata to add metadata to registered models
            if node_tags["hamilton.data_saver.sink"] == "mlflow":
                # skip if not registered model
                if "registered_model" not in result.keys():
                    return

                # get the registered model name (param of MLFlowModelSaver)
                model_name = result["registered_model"]["name"]
                version = result["registered_model"]["version"]
                materializer_node = self.graph[node_name]
                # get the "materialized node" defining the model
                materialized_node = self.graph[materializer_node.required_dependencies.pop()]
                # add the materialized node docstring as description
                # registered models have multiple versions
                self.client.update_registered_model(model_name, materialized_node.documentation)
                self.client.update_model_version(
                    model_name, version, materialized_node.documentation
                )

                # add node name as tag
                self.client.set_registered_model_tag(
                    model_name, key="node_name", value=materialized_node.name
                )
                self.client.set_model_version_tag(
                    model_name, version, key="node_name", value=materialized_node.name
                )

                # add origin function name as tag
                self.client.set_registered_model_tag(
                    model_name,
                    key="function_name",
                    value=materialized_node.originating_functions[0].__name__,
                )
                self.client.set_model_version_tag(
                    model_name,
                    version,
                    key="function_name",
                    value=materialized_node.originating_functions[0].__name__,
                )

                # add the materialized node @tag values as tags
                for k, v in materialized_node.tags.items():
                    # skip internal Hamilton tags
                    if "hamilton." in k:
                        continue
                    self.client.set_registered_model_tag(model_name, key=k, value=v)
                    self.client.set_model_version_tag(model_name, version, key=k, value=v)
                # TODO automatically collect model input signature; maybe simpler from user code

            # special case for matplotlib and plotly
            # log materialized figure. Allows great degree of control over rendering format
            # and also save interactive plotly visualization as HTML
            elif node_tags["hamilton.data_saver.sink"] in ["plt", "plotly"]:
                materializer_node = self.graph[node_name]
                materialized_node = self.graph[materializer_node.required_dependencies.pop()]
                figure = node_kwargs[materialized_node.name]

                path = get_path_from_metadata(result)
                if path:
                    self.client.log_figure(self.run_id, figure, path)
                else:
                    logger.warning(
                        f"Materialization result from node={node_name} has no recordable path: {result}. Materializer must have either "
                        f"'path' or 'file_metadata' keys."
                    )

            else:
                # log the materializer path as an artifact
                path = get_path_from_metadata(result)
                if path:
                    self.client.log_artifact(self.run_id, path, node_name)
                else:
                    logger.warning(
                        f"Materialization result from node={node_name} has no recordable path: {result}. Materializer must have either "
                        f"'path' or 'file_metadata' keys."
                    )
            return

        # log final_vars as artifacts
        if node_name not in self.final_vars:
            return

        # TODO refactor if/else as `handle_final_vars()`
        # log float and int as metrics
        if node_return_type in [float, int]:
            self.client.log_metric(self.run_id, key=node_name, value=float(result))

        # log str as text in .txt format
        elif isinstance(node_return_type, str):
            file_path = f"{node_name}.txt"
            with open(file_path, "w") as f:
                f.write(result)
            self.client.log_text(self.run_id, result, file_path)

        # log_dict (JSON) dictionary types; pickle if not json-serializable
        elif isinstance(node_return_type, dict):
            try:
                file_path = f"{node_name}.json"
                self.client.log_dict(self.run_id, result, file_path)
            # not json-serializable
            except TypeError:
                file_path = f"{node_name}.pickle"
                with open(file_path, "wb") as f:
                    pickle.dump(result, file=f)
                self.client.log_dict(self.run_id, result, file_path)

        # this puts less burden on users by not having to define materializers
        # for viz, but less control over rendering format
        elif node_return_type in FIGURE_TYPES:
            file_path = f"{node_name}.png"
            self.client.log_figure(self.run_id, result, file_path)

        # default to log_artifact in .pickle format
        else:
            file_path = f"{node_name}.pickle"
            with open(file_path, "wb") as f:
                pickle.dump(result, f)
            self.client.log_dict(self.run_id, result, file_path)

    def run_after_graph_execution(self, success: bool, *args, **kwargs):
        """End the MLFlow run"""
        # `status` is an enum value of mlflow.entities.RunStatus
        if success:
            self.client.set_terminated(self.run_id, status="FINISHED")
        else:
            self.client.set_terminated(self.run_id, status="FAILED")
        mlflow.end_run()

    def run_before_node_execution(self, *args, **kwargs):
        """Placeholder required to subclass NodeExecutionHook"""



from hamilton.plugins import h_ray


def test_parse_ray_remote_options_from_tags():

    tags = {
        f"{h_ray.RAY_REMOTE_TAG_NAMESPACE}.resources": '{"GPU": 1}',
        "another_tag": "another_value",
    }

    ray_options = h_ray.parse_ray_remote_options_from_tags(tags)

    assert len(ray_options) == 1
    assert "resources" in ray_options
    assert ray_options["resources"] == {"GPU": 1}



import abc
from abc import ABC
from types import ModuleType
from typing import TYPE_CHECKING, Any, Collection, Dict, List, Optional, Tuple, Type, final

from hamilton import graph_types, node

# This is only here for a type-hint
# As python types aren't real (they're determined at runtime), we can't have circular import resolved
# These are often necessary to handle typing -- as types don't have a perfect DAG of dependencies
# In this case, we're breaking the following loop:
#    -> lifecycle_api depends on graph_types and FunctionGraph
#    -> graph_types depends on hamilton.base
#    -> hamilton.base depends on lifecycle_api, as some interfaces for graph adapters live there
# To really fix this we should move everything user-facing out of base, which is a pretty sloppy name for a package anyway
# And put it where it belongs. For now we're OK with the TYPE_CHECKING hack
if TYPE_CHECKING:
    from hamilton.graph import FunctionGraph

from hamilton.graph_types import HamiltonGraph, HamiltonNode
from hamilton.lifecycle.base import (
    BaseDoBuildResult,
    BaseDoCheckEdgeTypesMatch,
    BaseDoNodeExecute,
    BaseDoValidateInput,
    BasePostGraphConstruct,
    BasePostGraphExecute,
    BasePostNodeExecute,
    BasePostTaskExecute,
    BasePreGraphExecute,
    BasePreNodeExecute,
    BasePreTaskExecute,
    BaseValidateGraph,
    BaseValidateNode,
)

try:
    from typing import override
except ImportError:
    override = lambda x: x  # noqa E731


class ResultBuilder(BaseDoBuildResult, abc.ABC):
    """Abstract class for building results. All result builders should inherit from this class and implement the build_result function.
    Note that applicable_input_type and output_type are optional, but recommended, for backwards
    compatibility. They let us type-check this. They will default to Any, which means that they'll
    connect to anything."""

    @abc.abstractmethod
    def build_result(self, **outputs: Any) -> Any:
        """Given a set of outputs, build the result.

        :param outputs: the outputs from the execution of the graph.
        :return: the result of the execution of the graph.
        """
        pass

    @override
    @final
    def do_build_result(self, outputs: Dict[str, Any]) -> Any:
        """Implements the do_build_result method from the BaseDoBuildResult class.
        This is kept from the user as the public-facing API is build_result, allowing us to change the
        API/implementation of the internal set of hooks"""
        return self.build_result(**outputs)

    def input_types(self) -> List[Type[Type]]:
        """Gives the applicable types to this result builder.
        This is optional for backwards compatibility, but is recommended.

        :return: A list of types that this can apply to.
        """
        return [Any]

    def output_type(self) -> Type:
        """Returns the output type of this result builder
        :return: the type that this creates
        """
        return Any


class LegacyResultMixin(ResultBuilder, ABC):
    """Backwards compatible legacy result builder. This utilizes a static method as we used to do that,
    although often times they got confused. If you want a result builder, use ResultBuilder above instead.
    """

    @staticmethod
    def build_result(**outputs: Any) -> Any:
        """Given a set of outputs, build the result.

        :param outputs: the outputs from the execution of the graph.
        :return: the result of the execution of the graph.
        """
        pass


class GraphAdapter(
    BaseDoNodeExecute,
    LegacyResultMixin,
    BaseDoValidateInput,
    BaseDoCheckEdgeTypesMatch,
    abc.ABC,
):
    """This is an implementation of HamiltonGraphAdapter, which has now been
    implemented with lifecycle methods/hooks."""

    @staticmethod
    @abc.abstractmethod
    def check_input_type(node_type: Type, input_value: Any) -> bool:
        """Used to check whether the user inputs match what the execution strategy & functions can handle.

        Static purely for legacy reasons.

        :param node_type: The type of the node.
        :param input_value: An actual value that we want to inspect matches our expectation.
        :return: True if the input is valid, False otherwise.
        """
        pass

    @staticmethod
    @abc.abstractmethod
    def check_node_type_equivalence(node_type: Type, input_type: Type) -> bool:
        """Used to check whether two types are equivalent.

        Static, purely for legacy reasons.

        This is used when the function graph is being created and we're statically type checking the annotations
        for compatibility.

        :param node_type: The type of the node.
        :param input_type: The type of the input that would flow into the node.
        :return: True if the types are equivalent, False otherwise.
        """
        pass

    @override
    @final
    def do_node_execute(
        self, run_id: str, node_: node.Node, kwargs: Dict[str, Any], task_id: Optional[str] = None
    ) -> Any:
        return self.execute_node(node_, kwargs)

    @override
    @final
    def do_validate_input(self, node_type: type, input_value: Any) -> bool:
        return self.check_input_type(node_type, input_value)

    @override
    @final
    def do_check_edge_types_match(self, type_from: type, type_to: type) -> bool:
        return self.check_node_type_equivalence(type_to, type_from)

    @abc.abstractmethod
    def execute_node(self, node: node.Node, kwargs: Dict[str, Any]) -> Any:
        """Given a node that represents a hamilton function, execute it.
        Note, in some adapters this might just return some type of "future".

        :param node: the Hamilton Node
        :param kwargs: the kwargs required to exercise the node function.
        :return: the result of exercising the node.
        """
        pass


class NodeExecutionHook(BasePreNodeExecute, BasePostNodeExecute, abc.ABC):
    """Implement this to hook into the node execution lifecycle. You can call anything before and after the driver"""

    @abc.abstractmethod
    def run_before_node_execution(
        self,
        *,
        node_name: str,
        node_tags: Dict[str, Any],
        node_kwargs: Dict[str, Any],
        node_return_type: type,
        task_id: Optional[str],
        run_id: str,
        node_input_types: Dict[str, Any],
        **future_kwargs: Any,
    ):
        """Hook that is executed prior to node execution.

        :param node_name: Name of the node.
        :param node_tags: Tags of the node
        :param node_kwargs: Keyword arguments to pass to the node
        :param node_return_type: Return type of the node
        :param task_id: The ID of the task, none if not in a task-based environment
        :param run_id: Run ID (unique in process scope) of the current run. Use this to track state.
        :param node_input_types: the input types to the node and what it is expecting
        :param future_kwargs: Additional keyword arguments -- this is kept for backwards compatibility
        """
        pass

    @override
    @final
    def pre_node_execute(
        self,
        *,
        run_id: str,
        node_: node.Node,
        kwargs: Dict[str, Any],
        task_id: Optional[str] = None,
    ):
        """Wraps the before_execution method, providing a bridge to an external-facing API. Do not override this!"""
        self.run_before_node_execution(
            node_name=node_.name,
            node_tags=node_.tags,
            node_kwargs=kwargs,
            node_return_type=node_.type,
            task_id=task_id,
            run_id=run_id,
            node_input_types={k: v[0] for k, v in node_.input_types.items()},
        )

    @abc.abstractmethod
    def run_after_node_execution(
        self,
        *,
        node_name: str,
        node_tags: Dict[str, Any],
        node_kwargs: Dict[str, Any],
        node_return_type: type,
        result: Any,
        error: Optional[Exception],
        success: bool,
        task_id: Optional[str],
        run_id: str,
        **future_kwargs: Any,
    ):
        """Hook that is executed post node execution.

        :param node_name: Name of the node in question
        :param node_tags: Tags of the node
        :param node_kwargs: Keyword arguments passed to the node
        :param node_return_type: Return type of the node
        :param result: Output of the node, None if an error occurred
        :param error: Error that occurred, None if no error occurred
        :param success: Whether the node executed successfully
        :param task_id: The ID of the task, none if not in a task-based environment
        :param run_id: Run ID (unique in process scope) of the current run. Use this to track state.
        :param future_kwargs: Additional keyword arguments -- this is kept for backwards compatibility
        """
        pass

    @override
    @final
    def post_node_execute(
        self,
        *,
        run_id: str,
        node_: node.Node,
        kwargs: Dict[str, Any],
        success: bool,
        error: Optional[Exception],
        result: Optional[Any],
        task_id: Optional[str] = None,
    ):
        """Wraps the after_execution method, providing a bridge to an external-facing API. Do not override this!"""
        self.run_after_node_execution(
            node_name=node_.name,
            node_tags=node_.tags,
            node_kwargs=kwargs,
            node_return_type=node_.type,
            result=result,
            error=error,
            task_id=task_id,
            success=success,
            run_id=run_id,
        )


class GraphExecutionHook(BasePreGraphExecute, BasePostGraphExecute):
    """Implement this to execute code before and after graph execution. This is useful for logging, etc..."""

    @override
    @final
    def post_graph_execute(
        self,
        *,
        run_id: str,
        graph: "FunctionGraph",
        success: bool,
        error: Optional[Exception],
        results: Optional[Dict[str, Any]],
    ):
        """Just delegates to the interface method, passing in the right data."""
        return self.run_after_graph_execution(
            graph=HamiltonGraph.from_graph(graph),
            success=success,
            error=error,
            results=results,
            run_id=run_id,
        )

    @override
    @final
    def pre_graph_execute(
        self,
        *,
        run_id: str,
        graph: "FunctionGraph",
        final_vars: List[str],
        inputs: Dict[str, Any],
        overrides: Dict[str, Any],
    ):
        """Implementation of the pre_graph_execute hook. This just converts the inputs to
        the format the user-facing hook is expecting -- performing a walk of the DAG to pass in
        the set of nodes to execute. Delegates to the interface method."""
        all_nodes, user_defined_nodes = graph.get_upstream_nodes(final_vars, inputs, overrides)
        nodes_to_execute = set(all_nodes) - set(user_defined_nodes)
        return self.run_before_graph_execution(
            graph=HamiltonGraph.from_graph(graph),
            final_vars=final_vars,
            inputs=inputs,
            overrides=overrides,
            execution_path=[item.name for item in nodes_to_execute],
            run_id=run_id,
        )

    @abc.abstractmethod
    def run_before_graph_execution(
        self,
        *,
        graph: graph_types.HamiltonGraph,
        final_vars: List[str],
        inputs: Dict[str, Any],
        overrides: Dict[str, Any],
        execution_path: Collection[str],
        run_id: str,
        **future_kwargs: Any,
    ):
        """This is run prior to graph execution. This allows you to do anything you want before the graph executes,
        knowing the basic information that was passed in.

        :param graph: Graph that is being executed
        :param final_vars: Output variables of the graph
        :param inputs: Input variables passed to the graph
        :param overrides: Overrides passed to the graph
        :param execution_path: Collection of nodes that will be executed --
            these are just the nodes (not input nodes) that will be run during the course of execution.
        :param run_id: Run ID (unique in process scope) of the current run. Use this to track state.
        :param future_kwargs: Additional keyword arguments -- this is kept for backwards compatibility
        """
        pass

    @abc.abstractmethod
    def run_after_graph_execution(
        self,
        *,
        graph: graph_types.HamiltonGraph,
        success: bool,
        error: Optional[Exception],
        results: Optional[Dict[str, Any]],
        run_id: str,
        **future_kwargs: Any,
    ):
        """This is run after graph execution. This allows you to do anything you want after the graph executes,
        knowing the results of the execution/any errors.

        :param graph: Graph that is being executed
        :param results: Results of the graph execution
        :param error: Error that occurred, None if no error occurred
        :param success: Whether the graph executed successfully
        :param run_id: Run ID (unique in process scope) of the current run. Use this to track state.
        :param future_kwargs: Additional keyword arguments -- this is kept for backwards compatibility
        """
        pass


class TaskExecutionHook(BasePreTaskExecute, BasePostTaskExecute, abc.ABC):
    """Implement this to run something after task execution. Tasks are tols used to group nodes.
    Note that this is currently run *inside* the task, although we do not guarantee where it will be run
    (it could easily move to outside the task)."""

    def pre_task_execute(
        self,
        *,
        run_id: str,
        task_id: str,
        nodes: List["node.Node"],
        inputs: Dict[str, Any],
        overrides: Dict[str, Any],
    ):
        self.run_before_task_execution(
            run_id=run_id,
            task_id=task_id,
            nodes=[HamiltonNode.from_node(n) for n in nodes],
            inputs=inputs,
            overrides=overrides,
        )

    def post_task_execute(
        self,
        *,
        run_id: str,
        task_id: str,
        nodes: List["node.Node"],
        results: Optional[Dict[str, Any]],
        success: bool,
        error: Exception,
    ):
        self.run_after_task_execution(
            run_id=run_id,
            task_id=task_id,
            nodes=[HamiltonNode.from_node(n) for n in nodes],
            results=results,
            success=success,
            error=error,
        )

    @abc.abstractmethod
    def run_before_task_execution(
        self,
        *,
        task_id: str,
        run_id: str,
        nodes: List[HamiltonNode],
        inputs: Dict[str, Any],
        overrides: Dict[str, Any],
        **future_kwargs,
    ):
        """Implement this to run something after task execution. Tasks are tols used to group nodes.
        Note that this is currently run *inside* the task, although we do not guarantee where it will be run
        (it could easily move to outside the task).

        :param task_id: ID of the task we're launching.
        :param run_id: ID of the run this is under.
        :param nodes: Nodes that are part of this task
        :param inputs: Inputs to the task
        :param overrides: Overrides passed to the task
        :param future_kwargs: Reserved for backwards compatibility.
        """
        pass

    @abc.abstractmethod
    def run_after_task_execution(
        self,
        *,
        task_id: str,
        run_id: str,
        nodes: List[HamiltonNode],
        results: Optional[Dict[str, Any]],
        success: bool,
        error: Exception,
        **future_kwargs,
    ):
        """Implement this to run something after task execution. See note in run_before_task_execution.

        :param task_id: ID of the task that was just executed
        :param run_id: ID of the run this was under.
        :param nodes: Nodes that were part of this task
        :param results: Results of the task, per-node
        :param success: Whether the task was successful
        :param error: The error the task threw, if any
        :param future_kwargs: Reserved for backwards compatibility.
        """
        pass


class EdgeConnectionHook(BaseDoCheckEdgeTypesMatch, BaseDoValidateInput, abc.ABC):
    """Implement this to customize edges that are allowed in the graph. You can do customizations around typing here."""

    @override
    @final
    def do_check_edge_types_match(self, *, type_from: type, type_to: type) -> bool:
        """Wraps the check_edge_types_match method, providing a bridge to an external-facing API. Do not override this!"""
        return self.check_edge_types_match(type_from, type_to)

    @abc.abstractmethod
    def check_edge_types_match(self, type_from: type, type_to: type, **kwargs: Any) -> bool:
        """This is run to check if edge types match. Note that this is an OR functionality
        -- this is run after we do some default checks, so this can only be permissive.
        Reach out if you want to be more restrictive than the default checks.

        :param type_from: The type of the node that is the source of the edge.
        :param type_to: The type of the node that is the destination of the edge.
        :param kwargs: This is kept for future backwards compatibility.
        :return: Whether or not the two node types form a valid edge.
        """
        pass

    @override
    @final
    def do_validate_input(self, *, node_type: type, input_value: Any) -> bool:
        """Wraps the validate_input method, providing a bridge to an external-facing API. Do not override this!"""
        return self.validate_input(node_type=node_type, input_value=input_value)

    @abc.abstractmethod
    def validate_input(self, node_type: type, input_value: Any, **kwargs: Any) -> bool:
        """This is run to check if the input is valid for the node type. Note that this is an OR functionality
        -- this is run after we do some default checks, so this can only be permissive.
        Reach out if you want to be more restrictive than the default checks.

        :param node_type: Type of the node that is accepting the input.
        :param input_value: Value of the input
        :param kwargs: Keyword arguments -- this is kept for future backwards compatibility.
        :return: Whether the input is valid for the node type.
        """
        pass


class NodeExecutionMethod(BaseDoNodeExecute):
    """API for executing a node. This takes in tags, callable, node name, and kwargs, and is
    responsible for executing the node and returning the result. Note this is not (currently)
    able to be layered together, although we may add that soon.
    """

    @override
    @final
    def do_node_execute(
        self,
        *,
        run_id: str,
        node_: node.Node,
        kwargs: Dict[str, Any],
        task_id: Optional[str] = None,
    ) -> Any:
        return self.run_to_execute_node(
            node_name=node_.name,
            node_tags=node_.tags,
            node_callable=node_.callable,
            node_kwargs=kwargs,
            task_id=task_id,
        )

    @abc.abstractmethod
    def run_to_execute_node(
        self,
        *,
        node_name: str,
        node_tags: Dict[str, Any],
        node_callable: Any,
        node_kwargs: Dict[str, Any],
        task_id: Optional[str],
        **future_kwargs: Any,
    ) -> Any:
        """This method is responsible for executing the node and returning the result.

        :param node_name: Name of the node.
        :param node_tags: Tags of the node.
        :param node_callable: Callable of the node.
        :param node_kwargs: Keyword arguments to pass to the node.
        :param task_id: The ID of the task, none if not in a task-based environment
        :param future_kwargs: Additional keyword arguments -- this is kept for backwards compatibility
        :return: The result of the node execution -- up to you to return this.
        """
        pass


class StaticValidator(BaseValidateGraph, BaseValidateNode):
    """Performs static validation of the DAG. Note that this has the option to perform default validation for each method --
    this means that if you don't implement one of these it is OK."""

    def run_to_validate_node(
        self, *, node: HamiltonNode, **future_kwargs
    ) -> Tuple[bool, Optional[str]]:
        """Override this to build custom node validations! Defaults to just returning that a node is valid so you don't have to implement it if you want to just implement a single method.
        Runs post node construction to validate a node. You have access to a bunch of metadata about the node, stored in the hamilton_node argument

        :param node: Node to validate
        :param future_kwargs: Additional keyword arguments -- this is kept for backwards compatibility
        :return: A tuple of whether the node is valid and an error
            message in the case of failure. Return [True, None] for a valid node.Otherwise, return a detailed error message -- this should have all context/debugging information, but does not need to
            mention the node name (it will be aggregated with others).
        """
        return True, None

    def run_to_validate_graph(
        self, graph: HamiltonGraph, **future_kwargs
    ) -> Tuple[bool, Optional[str]]:
        """Override this to build custom DAG validations! Default to just returning that the graph is valid, so you don't have to implement it if you want to just implement a single method.
        Runs post graph construction to validate a graph. You have access to a bunch of metadata about the graph, stored in the graph argument.

        :param graph: Graph to validate.
        :param future_kwargs: Additional keyword arguments -- this is kept for backwards compatibility
        :return: A tuple of whether the graph is valid and an error message in the case of failure. Return [True, None] for a valid graph.
            Otherwise, return a detailed error message -- this should have all context/debugging information.
        """
        return True, None

    @override
    @final
    def validate_node(self, *, created_node: node.Node) -> Tuple[bool, Optional[Exception]]:
        return self.run_to_validate_node(node=HamiltonNode.from_node(created_node))

    @override
    @final
    def validate_graph(
        self, *, graph: "FunctionGraph", modules: List[ModuleType], config: Dict[str, Any]
    ) -> Tuple[bool, Optional[Exception]]:
        return self.run_to_validate_graph(graph=HamiltonGraph.from_graph(graph))


class GraphConstructionHook(BasePostGraphConstruct, abc.ABC):
    """Hook that is run after graph construction. This allows you to register/capture info on the graph.
    Note that, in the case of materialization, this may be called multiple times (once when we create the graph,
    once when we materialize). Currently information into that is not exposed to the user, but we will be adding that in future
    iterations.
    """

    def post_graph_construct(
        self, *, graph: "FunctionGraph", modules: List[ModuleType], config: Dict[str, Any]
    ):
        self.run_after_graph_construction(graph=HamiltonGraph.from_graph(graph), config=config)

    @abc.abstractmethod
    def run_after_graph_construction(
        self, *, graph: HamiltonGraph, config: Dict[str, Any], **future_kwargs: Any
    ):
        """Hook that is run post graph construction. This allows you to register/capture info on the graph.
        A common pattern is to store something in your object's state here so that you can use it later
        (E.G. compute a hash on the graph)

        :param graph: Graph that was constructed
        :param config: Configuration used to construct the graph
        :param future_kwargs: Reserved for backwards compatibility.
        """
        pass



import pytest

from hamilton import function_modifiers
from hamilton.function_modifiers import base


def test_sanitize_function_name():
    assert base.sanitize_function_name("fn_name__v2") == "fn_name"
    assert base.sanitize_function_name("fn_name") == "fn_name"


def test_config_modifier_validate():
    def valid_fn() -> int:
        pass

    def valid_fn__this_is_also_valid() -> int:
        pass

    function_modifiers.config.when(key="value").validate(valid_fn__this_is_also_valid)
    function_modifiers.config.when(key="value").validate(valid_fn)

    def invalid_function__() -> int:
        pass

    with pytest.raises(base.InvalidDecoratorException):
        function_modifiers.config.when(key="value").validate(invalid_function__)


def test_config_when():
    def config_when_fn() -> int:
        pass

    annotation = function_modifiers.config.when(key="value")
    assert annotation.resolve(config_when_fn, {"key": "value"}) is not None
    assert annotation.resolve(config_when_fn, {"key": "wrong_value"}) is None


def test_config_when_not():
    def config_when_not_fn() -> int:
        pass

    annotation = function_modifiers.config.when_not(key="value")
    assert annotation.resolve(config_when_not_fn, {"key": "other_value"}) is not None
    assert annotation.resolve(config_when_not_fn, {"key": "value"}) is None


def test_config_when_in():
    def config_when_in_fn() -> int:
        pass

    annotation = function_modifiers.config.when_in(key=["valid_value", "another_valid_value"])
    assert annotation.resolve(config_when_in_fn, {"key": "valid_value"}) is not None
    assert annotation.resolve(config_when_in_fn, {"key": "another_valid_value"}) is not None
    assert annotation.resolve(config_when_in_fn, {"key": "not_a_valid_value"}) is None


def test_config_when_not_in():
    def config_when_not_in_fn() -> int:
        pass

    annotation = function_modifiers.config.when_not_in(
        key=["invalid_value", "another_invalid_value"]
    )
    assert annotation.resolve(config_when_not_in_fn, {"key": "invalid_value"}) is None
    assert annotation.resolve(config_when_not_in_fn, {"key": "another_invalid_value"}) is None
    assert annotation.resolve(config_when_not_in_fn, {"key": "valid_value"}) is not None


def test_config_name_resolution():
    def fn__v2() -> int:
        pass

    annotation = function_modifiers.config.when(key="value")
    result = annotation.resolve(fn__v2, {"key": "value"})
    assert result.__name__ == "fn"
    assert result.__original_name__ == "fn__v2"


def test_config_when_with_custom_name():
    def config_when_fn() -> int:
        pass

    annotation = function_modifiers.config.when(key="value", name="new_function_name")
    result = annotation.resolve(config_when_fn, {"key": "value"})
    assert result.__name__ == "new_function_name"
    assert result.__original_name__ == "config_when_fn"


def test_config_base_resolve_nodes():
    def config_fn() -> int:
        pass

    annotation = function_modifiers.config(lambda conf: conf["key"] == "value")
    assert annotation.resolve(config_fn, {"key": "value"}) is not None


def test_config_base_resolve_nodes_no_resolve():
    def config_fn() -> int:
        pass

    annotation = function_modifiers.config(lambda conf: conf.get("key") == "value")
    assert annotation.resolve(config_fn, {}) is None


def test_config_base_resolve_nodes_end_to_end():
    def config_fn() -> int:
        pass

    # tests that the full resolver works
    annotation = function_modifiers.config(lambda conf: conf.get("key") == "value")
    config_fn = annotation(config_fn)
    nodes = base.resolve_nodes(config_fn, {})
    assert len(nodes) == 0



from hamilton.function_modifiers import parameterize_sources, parametrized_input


def input_1() -> int:
    return 1


def input_2() -> int:
    return 2


def input_3() -> int:
    return 3


@parametrized_input(
    parameter="input_value_tbd",
    variable_inputs={
        "input_1": ("output_1", "function_with_multiple_inputs called using input_1"),
        "input_2": ("output_2", "function_with_multiple_inputs called using input_2"),
    },
)
def function_with_multiple_inputs(input_value_tbd: int, static_value: int) -> int:
    return input_value_tbd + static_value


# We don't prefer this style of specifying the values. i.e. kwarg with {}.
@parameterize_sources(output_12={"input_value_tbd1": "input_1", "input_value_tbd2": "input_2"})
def function_with_two_parameters(
    input_value_tbd1: int, input_value_tbd2: int, static_value: int
) -> int:
    """function_with_multiple_inputs called using {input_value_tbd1} and {input_value_tbd2}

    :return: creates {output_name}
    """
    return input_value_tbd1 + input_value_tbd2 + static_value


# We prefer this style of specifying the values. i.e. kwarg with dict().
@parameterize_sources(
    output_123=dict(
        input_value_tbd1="input_1", input_value_tbd2="input_2", input_value_tbd3="input_3"
    )
)
def function_with_three_parameters(
    input_value_tbd1: int, input_value_tbd2: int, input_value_tbd3: int, static_value: int
) -> int:
    """function_with_multiple_inputs called using {input_value_tbd1} and {input_value_tbd2} and {input_value_tbd3}

    :param {input_value_tbd1}:
    :param {input_value_tbd2}:
    :param {input_value_tbd3}:
    :param static_value:
    :return: {output_name}
    """
    return input_value_tbd1 + input_value_tbd2 + input_value_tbd3 + static_value



from typing import Any

try:
    import geopandas as gpd
except ImportError:
    raise NotImplementedError("geopandas is not installed.")

from hamilton import registry

DATAFRAME_TYPE = gpd.GeoDataFrame
COLUMN_TYPE = gpd.GeoSeries


@registry.get_column.register(gpd.GeoDataFrame)
def get_column_geopandas(df: gpd.GeoDataFrame, column_name: str) -> gpd.GeoSeries:
    return df[column_name]


@registry.fill_with_scalar.register(gpd.GeoDataFrame)
def fill_with_scalar_geopandas(
    df: gpd.GeoDataFrame, column_name: str, value: Any
) -> gpd.GeoDataFrame:
    df[column_name] = value
    return df


def register_types():
    """Function to register the types for this extension."""
    registry.register_types("geopandas", DATAFRAME_TYPE, COLUMN_TYPE)


register_types()



import dataclasses
import functools
import inspect
from types import ModuleType
from typing import Any, Callable, Dict, List, Optional, Tuple

from hamilton import node
from hamilton.graph import FunctionGraph
from hamilton.lifecycle.base import (
    BaseDoBuildResult,
    BaseDoNodeExecute,
    BaseDoValidateInput,
    BasePostGraphConstruct,
    BasePostGraphExecute,
    BasePostNodeExecute,
    BasePostTaskExecute,
    BasePreDoAnythingHook,
    BasePreGraphExecute,
    BasePreNodeExecute,
    BasePreTaskExecute,
    BaseValidateGraph,
    BaseValidateNode,
    LifecycleAdapterSet,
)
from hamilton.node import Node


@dataclasses.dataclass
class HookCall:
    sequence_number: int
    name: str
    fn: Callable
    bound_kwargs: Dict[str, Any]
    result: Any


class SentinelException(Exception):
    pass


# Allows us to track the order in which multiple lifecycle hooks are called
# This ensures that we can get order of call across operations
# We'll want to test for order, not absolute position, as it is global
sequence_number = 0


# This is slightly hacky, but its a quick way to decorate the lifecycle hooks
class ExtendToTrackCalls:
    def __init__(self, name: str):
        # We just pass in self, cause why not?
        adapter_set = LifecycleAdapterSet(self)
        for lifecycle_step, _ in {
            **adapter_set.sync_methods,
            **adapter_set.async_methods,
            **adapter_set.sync_hooks,
            **adapter_set.async_hooks,
            **adapter_set.sync_validators,
        }.items():
            # We know there's just one in this case
            setattr(self, lifecycle_step, self._wrap_fn(getattr(self, lifecycle_step)))
        self._calls = []
        self._name = name

    def _wrap_fn(self, fn: Callable):
        @functools.wraps(fn)
        def wrapped(*args, **kwargs):
            global sequence_number
            sequence_number += 1
            sig = inspect.signature(fn)
            bound = sig.bind(*args, **kwargs)
            out = fn(*args, **kwargs)
            self._calls.append(HookCall(sequence_number, self._name, fn, bound.arguments, out))
            return out

        return wrapped

    @property
    def calls(self) -> List[HookCall]:
        return self._calls

    @property
    def name(self):
        return self._name


class TrackingPreDoAnythingHook(BasePreDoAnythingHook, ExtendToTrackCalls):
    def pre_do_anything(self):
        pass


class TrackingPostGraphConstructHook(ExtendToTrackCalls, BasePostGraphConstruct):
    def post_graph_construct(
        self, graph: FunctionGraph, modules: List[ModuleType], config: Dict[str, Any]
    ):
        pass


class TrackingPreGraphExecuteHook(ExtendToTrackCalls, BasePreGraphExecute):
    def pre_graph_execute(
        self,
        run_id: str,
        graph: FunctionGraph,
        final_vars: List[str],
        inputs: Dict[str, Any],
        overrides: Dict[str, Any],
    ):
        pass


class TrackingPreTaskExecuteHook(ExtendToTrackCalls, BasePreTaskExecute):
    def pre_task_execute(
        self,
        run_id: str,
        task_id: str,
        nodes: List[node.Node],
        inputs: Dict[str, Any],
        overrides: Dict[str, Any],
    ):
        pass


class TrackingPreNodeExecuteHook(ExtendToTrackCalls, BasePreNodeExecute):
    def pre_node_execute(
        self, run_id: str, node_: Node, kwargs: Dict[str, Any], task_id: Optional[str] = None
    ):
        pass


class TrackingPostNodeExecuteHook(ExtendToTrackCalls, BasePostNodeExecute):
    def post_node_execute(
        self,
        run_id: str,
        node_: Node,
        kwargs: Dict[str, Any],
        success: bool,
        error: Optional[Exception],
        result: Any,
        task_id: Optional[str] = None,
    ):
        pass


class TrackingPostTaskExecuteHook(ExtendToTrackCalls, BasePostTaskExecute):
    def post_task_execute(
        self,
        run_id: str,
        task_id: str,
        nodes: List[node.Node],
        results: Optional[Dict[str, Any]],
        success: bool,
        error: Exception,
    ):
        pass


class TrackingPostGraphExecuteHook(ExtendToTrackCalls, BasePostGraphExecute):
    def post_graph_execute(
        self,
        run_id: str,
        graph: FunctionGraph,
        success: bool,
        error: Optional[Exception],
        results: Optional[Dict[str, Any]],
    ):
        pass


class TrackingDoValidateInputMethod(ExtendToTrackCalls, BaseDoValidateInput):
    def __init__(self, name: str, valid: bool = True):
        super().__init__(name)
        self._valid = valid

    def do_validate_input(self, node_type: type, input_value: Any) -> bool:
        return self._valid


class TrackingDoNodeExecuteHook(ExtendToTrackCalls, BaseDoNodeExecute):
    def __init__(self, name: str, additional_value: int):
        super().__init__(name)
        self._additional_value = additional_value

    def do_node_execute(
        self, run_id: str, node_: node.Node, kwargs: Dict[str, Any], task_id: Optional[str] = None
    ) -> Any:
        if node_.type == int and node_.name != "n_iters":
            return node_(**kwargs) + self._additional_value
        return node_(**kwargs)


class TrackingDoBuildResultMethod(ExtendToTrackCalls, BaseDoBuildResult):
    def __init__(self, name: str, result: Any):
        super().__init__(name)
        self._result = result

    def do_build_result(self, outputs: Dict[str, Any]) -> Any:
        return self._result


class TrackingValidateNodeValidator(ExtendToTrackCalls, BaseValidateNode):
    def __init__(self, name: str, valid: bool, message: Optional[str]):
        super().__init__(name)
        self._valid = valid
        self._message = message

    def validate_node(self, *, created_node: node.Node) -> Tuple[bool, Optional[str]]:
        return self._valid, self._message


class TrackingValidateGraphValidator(ExtendToTrackCalls, BaseValidateGraph):
    def __init__(self, name: str, valid: bool, message: Optional[str]):
        super().__init__(name)
        self._valid = valid
        self._message = message

    def validate_graph(
        self, *, graph: "FunctionGraph", modules: List[ModuleType], config: Dict[str, Any]
    ) -> Tuple[bool, Optional[str]]:
        return self._valid, self._message



import collections
import functools

from hamilton.htypes import Collect, Parallelizable

_fn_call_counter = collections.Counter()


def _track_fn_call(fn) -> callable:
    @functools.wraps(fn)
    def wrapped(*args, **kwargs):
        _fn_call_counter[fn.__name__] += 1
        return fn(*args, **kwargs)

    return wrapped


def _reset_counter():
    _fn_call_counter.clear()


@_track_fn_call
def not_to_repeat() -> int:
    return -1


@_track_fn_call
def number_to_repeat(iterations: int) -> Parallelizable[int]:
    for i in range(iterations):
        yield i


@_track_fn_call
def something_else_not_to_repeat() -> int:
    return -2


@_track_fn_call
def double(number_to_repeat: int) -> int:
    return number_to_repeat * 2


@_track_fn_call
def summed(double: Collect[int], not_to_repeat: int, something_else_not_to_repeat: int) -> int:
    return sum(double) + not_to_repeat + something_else_not_to_repeat



from typing import Any, Type

import pandera as pa

from hamilton import registry
from hamilton.data_quality import base
from hamilton.htypes import custom_subclass_check

pandera_supported_extensions = frozenset(["pandas", "dask", "pyspark_pandas"])


class PanderaDataFrameValidator(base.BaseDefaultValidator):
    """Pandera schema validator for dataframes"""

    def __init__(self, schema: pa.DataFrameSchema, importance: str):
        super(PanderaDataFrameValidator, self).__init__(importance)
        self.schema = schema

    @classmethod
    def applies_to(cls, datatype: Type[Type]) -> bool:
        for extension_name in pandera_supported_extensions:
            if extension_name in registry.DF_TYPE_AND_COLUMN_TYPES:
                df_type = registry.DF_TYPE_AND_COLUMN_TYPES[extension_name][registry.DATAFRAME_TYPE]
                result = custom_subclass_check(datatype, df_type)
                if result:
                    return True
        return False

    def description(self) -> str:
        return "Validates that the returned dataframe matches the pander"

    def validate(self, data: Any) -> base.ValidationResult:
        try:
            result = self.schema.validate(data, lazy=True, inplace=True)
            if hasattr(result, "dask"):
                result.compute()
        except pa.errors.SchemaErrors as e:
            return base.ValidationResult(
                passes=False,
                message=str(e),
                diagnostics={"schema_errors": e.schema_errors},
            )
        return base.ValidationResult(
            passes=True,
            message=f"Data passes pandera check for schema {str(self.schema)}",
            # TDOO -- add diagnostics data with serialized the schema
        )

    @classmethod
    def arg(cls) -> str:
        return "schema"  # TODO -- determine whether we want to allow other schemas

    @classmethod
    def name(cls) -> str:
        return "pandera_schema_validator"


class PanderaSeriesSchemaValidator(base.BaseDefaultValidator):
    """Pandera schema validator for series"""

    def __init__(self, schema: pa.SeriesSchema, importance: str):
        super(PanderaSeriesSchemaValidator, self).__init__(importance)
        self.schema = schema

    @classmethod
    def applies_to(cls, datatype: Type[Type]) -> bool:
        for extension_name in pandera_supported_extensions:
            if extension_name in registry.DF_TYPE_AND_COLUMN_TYPES:
                df_type = registry.DF_TYPE_AND_COLUMN_TYPES[extension_name][registry.COLUMN_TYPE]
                result = custom_subclass_check(datatype, df_type)
                if result:
                    return True
        return False

    def description(self) -> str:
        pass

    def validate(self, data: Any) -> base.ValidationResult:
        try:
            result = self.schema.validate(data, lazy=True, inplace=True)
            if hasattr(result, "dask"):
                result.compute()
        except pa.errors.SchemaErrors as e:
            return base.ValidationResult(
                passes=False,
                message=str(e),
                diagnostics={"schema_errors": e.schema_errors},
            )
        return base.ValidationResult(
            passes=True,
            message=f"Data passes pandera check for schema {str(self.schema)}",
            # TDOO -- add diagnostics data with serialized the schema
        )

    @classmethod
    def arg(cls) -> str:
        return "schema"  # TODO -- determine whether we want to allow other schemas

    @classmethod
    def name(cls) -> str:
        return "pandera_schema_validator"


PANDERA_VALIDATORS = [PanderaDataFrameValidator, PanderaSeriesSchemaValidator]



import argparse
import os
from pathlib import Path

from hamilton import telemetry
from hamilton.plugins.h_experiments.cache import JsonCache


def main():
    try:
        import fastapi  # noqa: F401
        import fastui  # noqa: F401
        import uvicorn
    except ModuleNotFoundError:
        raise ModuleNotFoundError(
            "Some dependencies are missing. Make sure to `pip install sf-hamilton[experiments]`"
        )
    if telemetry.is_telemetry_enabled():
        telemetry.create_and_send_expt_server_event("startup")
    parser = argparse.ArgumentParser(prog="hamilton-experiments")
    parser.description = "Hamilton Experiment Server launcher"

    parser.add_argument(
        "path",
        metavar="path",
        type=str,
        default="./experiments",
        nargs="?",
        help="Set HAMILTON_EXPERIMENTS_PATH environment variable",
    )
    parser.add_argument("--host", default="127.0.0.1", type=str, help="Bind to this address")
    parser.add_argument("--port", default=8123, type=int, help="Bind to this port")

    args = parser.parse_args()

    try:
        JsonCache(args.path)
    except Exception as e:
        raise ValueError(f"Server failed to launch. No metadata cache found at {args.path}") from e

    # set environment variable that FastAPI will use
    os.environ["HAMILTON_EXPERIMENTS_PATH"] = str(Path(args.path).resolve())

    uvicorn.run("hamilton.plugins.h_experiments.server:app", host=args.host, port=args.port)
    if telemetry.is_telemetry_enabled():
        telemetry.create_and_send_expt_server_event("shutdown")


if __name__ == "__main__":
    main()



"""
This module houses functions and tools to interact with off-the-shelf
dataflows.

TODO: expect this to have a CLI interface in the future.
"""

import functools
import importlib
import json
import logging
import os
import shutil
import sys
import time
import urllib.error
import urllib.request
from types import ModuleType
from typing import Callable, Dict, List, NamedTuple, Optional, Tuple, Type, Union

from hamilton import driver, telemetry

logger = logging.getLogger(__name__)

"""
Paths that we care about.
Assumptions:
 - ~/.hamilton/dataflows/ is where we store dataflows. This is fine because we should be able to save locally.
 - This will act as a cache for dataflows.
 - People can modify this location globally.
 - Directory should mirror github (e.g. contrib/hamilton/contrib/user/{user}/{dataflow})
 - We don't have to deal with the python path because this module knows how to import it.

TODOs:
 - finish init
 - make sure it works on windows
 - make functions more robust (see comments in functions).
"""
COMMON_PATH = "{commit_ish}/contrib/hamilton/contrib"
BASE_URL = f"https://raw.githubusercontent.com/dagworks-inc/hamilton/{COMMON_PATH}"
DATAFLOW_FOLDER = os.path.expanduser("~/.hamilton/dataflows")
USER_PATH = DATAFLOW_FOLDER + "/" + COMMON_PATH + "/user/{user}/{dataflow}"
OFFICIAL_PATH = DATAFLOW_FOLDER + "/" + COMMON_PATH + "/dagworks/{dataflow}"


def _track_function_call(call_fn: Callable) -> Callable:
    """Decorator to wrap the __call__ to count usage.

    :param call_fn: the `__call__` function.
    :return: the wrapped call function.
    """

    @functools.wraps(call_fn)
    def track_call(*args, **kwargs):
        event_json = telemetry.create_dataflow_function_invocation_event_json(call_fn.__name__)
        telemetry.send_event_json(event_json)
        return call_fn(*args, **kwargs)

    return track_call


def _track_download(is_official: bool, user: Optional[str], dataflow_name: str, version: str):
    """Inner function to track "downloads" of a dataflow.

    :param is_official: is this an official dataflow? False == user.
    :param user: If not official, what is the github user name.
    :param dataflow_name: the name of the dataflow
    :param version: the version. Either git hash, or the package version.
    """
    if is_official:
        category = "DAGWORKS"
    else:
        category = "USER"
    event_json = telemetry.create_dataflow_download_event_json(
        category, user, dataflow_name, version
    )
    telemetry.send_event_json(event_json)


def _get_request(url: str) -> Tuple[int, str]:
    """Makes a GET request to the given URL and returns the status code and response data.

    :param url: the url to make the request to.
    :return: tuple of status code and text response decoded as utf-8.
    """
    try:
        with urllib.request.urlopen(url) as response:
            data = response.read().decode("utf-8")
            return response.status, data

    except urllib.error.HTTPError as e:
        return e.code, e.reason


@_track_function_call
def clear_storage():
    """Clears all the data under DATAFLOW_FOLDER. By default its "~/.hamilton/dataflows"."""
    if os.path.exists(DATAFLOW_FOLDER):
        shutil.rmtree(DATAFLOW_FOLDER)
        logger.info(f"Cleared storage at {DATAFLOW_FOLDER}")
    else:
        logger.info(f"Folder {DATAFLOW_FOLDER} does not exist.")


# want TTL cache on this call to not get rate limited.
last_time_called = None
last_resolve_value = None


@_track_function_call
def resolve_latest_branch_commit(branch: str = "main") -> str:
    """Resolves https://api.github.com/repos/DAGWorks-Inc/hamilton/git/refs/heads/{branch}

    :return: commit for what "main" is.
    """
    global last_time_called
    global last_resolve_value
    if last_time_called is not None and last_resolve_value is not None:
        if (last_time_called + 60.0) > time.time():
            logger.info("using cached resolve_latest")
            return last_resolve_value
    url = f"https://api.github.com/repos/DAGWorks-Inc/hamilton/git/refs/heads/{branch}"
    status_code, text = _get_request(url)
    # response = requests.get(url)
    if status_code != 200:
        raise ValueError(
            f"Failed to resolve latest commit for branch {branch}:\n{status_code}\n{text}"
        )
    commit_sha = json.loads(text)["object"]["sha"]
    last_time_called = time.time()
    last_resolve_value = commit_sha
    return commit_sha


@_track_function_call
def latest_commit(dataflow: str, user: str = None) -> str:
    """Determines the latest commit for a dataflow.

    This is useful to know if you want to pull the latest version of a dataflow.

    :param dataflow: the string name of the dataflow
    :param user: the name of the user. None if official.
    :return: the commit sha.
    """
    if user:
        url = f"https://hub.dagworks.io/commits/Users/{user}/{dataflow}/commit.txt"
    else:
        url = f"https://hub.dagworks.io/commits/DAGWorks/{dataflow}/commit.txt"
    status_code, text = _get_request(url)
    if status_code != 200:
        raise ValueError(
            f"Failed to resolve latest commit [{url}] for {user}/{dataflow}:\n{status_code}\n{text}"
        )
    chunk_index = text.find("[commit::")
    if chunk_index < 0:
        raise ValueError("No commit found")
    # Gets the latest commit from potentially multiple.
    commit = text[chunk_index + len("[commit::") :]
    commit_sha = commit[: commit.find("]")]
    return commit_sha


@_track_function_call
def pull_module(dataflow: str, user: str = None, version: str = "latest", overwrite: bool = False):
    """Pulls a dataflow module.

    Saves to hamilton.dataflow.USER_PATH. An import should just work right after doing this.

    It performs the following:

      1. Creates a URL to pull from github.
      2. Pulls the code for the dataflow.
      3. Save to the local location based on hamilton.dataflow.USER_PATH.

    :param dataflow: the dataflow name.
    :param user: the user's github handle.
    :param version: the commit version. "latest" will resolve to the most recent commit, else pass \
        a commit SHA.
    :param overwrite: whether to overwrite. Default is False.
    """
    if version == "latest":
        version = latest_commit(dataflow, user)
    if user:
        _track_download(False, user, dataflow, version)
        logger.info(f"pulling dataflow {user}/{dataflow} with version {version}")
        local_file_path = USER_PATH.format(commit_ish=version, user=user, dataflow=dataflow)
    else:
        _track_download(True, None, dataflow, version)
        logger.info(f"pulling official dataflow {dataflow} with version {version}")
        local_file_path = OFFICIAL_PATH.format(commit_ish=version, dataflow=dataflow)

    h_files = [
        "__init__.py",
        "requirements.txt",
        "README.md",
        "valid_configs.jsonl",
        "tags.json",
    ]

    if os.path.exists(local_file_path) and not overwrite:
        raise ValueError(
            f"Dataflow {user}/{dataflow} with version {version} already exists locally. Not downloading."
        )
    os.makedirs(local_file_path, exist_ok=True)
    for h_file in h_files:
        if user:
            url = BASE_URL.format(commit_ish=version) + f"/user/{user}/{dataflow}/{h_file}"
        else:
            url = BASE_URL.format(commit_ish=version) + f"/dagworks/{dataflow}/{h_file}"
        # response = requests.get(url)
        status_code, text = _get_request(url)
        if status_code == 404:
            raise ValueError(f"Dataflow {user}/{dataflow}/{h_file} not found at {url}.")
        elif status_code != 200:
            raise ValueError(
                f"Dataflow {user}/{dataflow} returned status code {status_code}:\n{text}"
            )
        file_contents = text
        if os.path.exists(os.path.join(local_file_path, h_file)):
            logger.info(f"Warning: overwriting {h_file} in {local_file_path}")
        with open(os.path.join(local_file_path, h_file), "w") as f:
            f.write(file_contents)
            logger.info(f"wrote {h_file} to {local_file_path}")


@_track_function_call
def import_module(
    dataflow: str, user: str = None, version: str = "latest", overwrite: bool = False
) -> ModuleType:
    """Pulls & imports dataflow code from github and returns a module.

    .. code-block:: python

        from hamilton import dataflows, driver
        # downloads into ~/.hamilton/dataflows and loads the module -- WARNING: ensure you know what code you're importing!
        # NAME_OF_DATAFLOW = dataflow.import_module("NAME_OF_DATAFLOW") # if using official dataflow
        NAME_OF_DATAFLOW = dataflow.import_module("NAME_OF_DATAFLOW", "NAME_OF_USER")
        dr = (
          driver.Builder()
          .with_config({})  # replace with configuration as appropriate
          .with_modules(NAME_OF_DATAFLOW)
          .build()
        )
        # execute the dataflow, specifying what you want back. Will return a dictionary.
        result = dr.execute(
          [NAME_OF_DATAFLOW.FUNCTION_NAME, ...],  # this specifies what you want back
          inputs={...}  # pass in inputs as appropriate
        )

    :param dataflow: the name of the dataflow.
    :param user: Optional. If none it assumes official.
    :param version: the version to get. "latest" will resolve to the most recent commit. \
        Otherwise pass a the commit SHA you want to pull.
    :param overwrite: whether to overwrite the local path. Default is False.
    :return: a Module that you can then pass to Hamilton.
    """
    if version == "latest":
        version = latest_commit(dataflow, user)
    if user:
        local_file_path = (
            USER_PATH.format(commit_ish=version, user=user, dataflow=dataflow) + "/__init__.py"
        )
    else:
        local_file_path = (
            OFFICIAL_PATH.format(commit_ish=version, dataflow=dataflow) + "/__init__.py"
        )
    if not os.path.exists(local_file_path) or overwrite:
        pull_module(dataflow, user, version=version, overwrite=overwrite)
    else:
        logger.info(f"Found {user}/{dataflow} with version {version} locally. Not downloading.")

    if dataflow in sys.modules:
        logger.info(
            f"Warning: overwriting existing {dataflow} module with version {sys.modules[dataflow].__version__}"
        )
    spec = importlib.util.spec_from_file_location(dataflow, local_file_path)
    module = importlib.util.module_from_spec(spec)
    module.__version__ = version
    sys.modules[dataflow] = module
    spec.loader.exec_module(module)
    return module


class InspectResult(NamedTuple):
    version: str  # git commit sha/package version
    user: str  # github user URL
    dataflow: str  # dataflow URL
    python_dependencies: List[str]  # python dependencies
    configurations: List[str]  # configurations for the dataflow stored as a JSON string


@_track_function_call
def inspect(dataflow: str, user: str = None, version: str = "latest") -> InspectResult:
    """Inspects a dataflow for information.

    This is a helper function to get information about a dataflow that exists locally.
    It does not get more information because we don't want to assume we can import the module.

    .. code-block:: python

        from hamilton import dataflows

        info = dataflows.inspect("text_summarization", "zilto")

    :param dataflow: the dataflow name.
    :param user: the github name of the user. None for DAGWorks official.
    :param version: the version to inspect. "latest" will resolve to the most recent commit, else pass \
       a commit SHA.
    :return: hamilton.dataflow.InspectResult object that contains version, user URL, dataflow URL, python dependencies, configurations.
    """
    if version == "latest":
        version = latest_commit(dataflow, user)

    if user:
        local_file_path = USER_PATH.format(commit_ish=version, user=user, dataflow=dataflow)
        dataflow_url = (
            f"https://github.com/dagworks-inc/hamilton/tree/{version}/contrib/contrib/user/{user}/{dataflow}",
        )
        user_url = f"https://github.com/{user}"
    else:
        local_file_path = OFFICIAL_PATH.format(commit_ish=version, dataflow=dataflow)
        dataflow_url = (
            f"https://github.com/dagworks-inc/hamilton/tree/{version}/contrib/contrib/dagworks/{dataflow}",
        )
        user_url = None
    if not os.path.exists(local_file_path):
        raise ValueError(
            f"Dataflow {user or 'dagworks'}/{dataflow} with version {version} does not exist locally. Not inspecting."
        )
    # return dictionary of python deps, inputs, nodes, designated outputs, commit hash
    info: Dict[str, Union[str, List[Dict], List[str]]] = {
        "version": version,
        "user": user_url,
        "dataflow": dataflow_url,
        "python_dependencies": [],
        "configurations": [],
    }

    with open(os.path.join(local_file_path, "requirements.txt"), "r") as f:
        file_contents = [line.strip() for line in f.readlines()]
        info["python_dependencies"] = file_contents

    with open(os.path.join(local_file_path, "valid_configs.jsonl"), "r") as f:
        file_contents = [line.strip() for line in f.readlines()]
        info["configurations"] = file_contents
    return InspectResult(**info)


class InspectModuleResult(NamedTuple):
    version: str  # git commit sha/package version
    user: str  # github user URL
    dataflow: str  # dataflow URL
    python_dependencies: List[str]  # python dependencies
    configurations: List[str]  # configurations for the dataflow stored as a JSON string
    possible_inputs: List[Tuple[str, Type]]
    nodes: List[Tuple[str, Type]]
    designated_outputs: List[Tuple[str, Type]]


@_track_function_call
def inspect_module(module: ModuleType) -> InspectModuleResult:
    """Inspects the import module for information.

    This does more than `inspect` because the module has been loaded and thus
    we can put it into a Hamilton driver and ask questions of it.

    .. code-block:: python

        from hamilton.contrib.user.zilto import text_summarization
        from hamilton import dataflows

        info = dataflows.inspect_module(text_summarization)

    :param module: the module with Hamilton code to deeply introspect.
    :return: hamilton.dataflow.InspectModuleResult object.
    """
    if not module.__file__.startswith(DATAFLOW_FOLDER):
        logger.info("not a downloaded dataflow module.")
        return None
    version = module.__version__
    dataflow = module.__file__.split("/")[-2]
    if "hamilton/contrib/dagworks" in module.__file__:
        user = None
        local_file_path = OFFICIAL_PATH.format(commit_ish=version, dataflow=dataflow)
        user_url = None
        dataflow_url = f"https://github.com/dagworks-inc/hamilton/tree/{version}/contrib/contrib/dagworks/{dataflow}"
    else:
        user = module.__file__.split("/")[-3]
        user_url = f"https://github.com/{user}"
        local_file_path = USER_PATH.format(commit_ish=version, user=user, dataflow=dataflow)
        dataflow_url = f"https://github.com/dagworks-inc/hamilton/tree/{version}/contrib/contrib/user/{user}/{dataflow}"

    if not os.path.exists(local_file_path):
        raise ValueError(
            f"Dataflow {user or 'dagworks'}/{dataflow} with version {version} does not exist locally. Not inspecting."
        )
    # return dictionary of python deps, inputs, nodes, designated outputs, commit hash
    info: Dict[str, Union[str, List[Dict], List[str], List[Tuple[str, Type]]]] = {
        "version": version,
        "user": user_url,
        "dataflow": dataflow_url,
        "python_dependencies": [],
        "possible_inputs": [],
        "nodes": [],
        "designated_outputs": [],
        "configurations": [],
    }

    with open(os.path.join(local_file_path, "requirements.txt"), "r") as f:
        file_contents = f.readlines()
        info["python_dependencies"] = file_contents

    with open(os.path.join(local_file_path, "valid_configs.jsonl"), "r") as f:
        file_contents = [json.loads(s) for s in f.readlines()]
        info["configurations"] = file_contents

    dr = driver.Driver(info["configurations"][0], module)
    vars = dr.list_available_variables()
    info["possible_inputs"] = [(var.name, var.type) for var in vars if var.is_external_input]
    info["nodes"] = [(var.name, var.type) for var in vars if not var.is_external_input]
    info["designated_outputs"] = [
        (var.name, var.type) for var in vars if var.tags.get("designated_output", False) == "True"
    ]
    return InspectModuleResult(**info)


@_track_function_call
def install_dependencies_string(dataflow: str, user: str = None, version: str = "latest") -> str:
    """Returns a string for the user to install dependencies.

    :param dataflow: the name of the dataflow.
    :param user: the github name of the user.
    :param version: the version to inspect. "latest" will resolve to the most recent commit, else pass \
         a commit SHA.
    :return: pip install string to use.
    """
    if version == "latest":
        version = latest_commit(dataflow, user)
    if user:
        local_file_path = USER_PATH.format(commit_ish=version, user=user, dataflow=dataflow)
    else:
        local_file_path = OFFICIAL_PATH.format(commit_ish=version, dataflow=dataflow)
    if not os.path.exists(local_file_path):
        logger.info(
            "Dataflow does not exist locally. Can't provide details on how to install dependencies."
        )
        return ""
    return f"pip install -r {local_file_path}/requirements.txt"


from importlib.metadata import PackageNotFoundError
from importlib.metadata import version as pkg_version


@_track_function_call
def are_py_dependencies_satisfied(dataflow, user=None, version="latest"):
    """Given a commit_ish, user & dataflow, reads the requirements.txt and checks if \
    those dependencies have been satisfied in the currently running python interpreter.

    Note: this does not handle versions. Just whether the package is installed or not.

    :param dataflow: the name of the dataflow.
    :param user: the github name of the user. None if DAGWorks official.
    :param version: the version to inspect. "latest" will resolve to the most recent commit, else pass \
         a commit SHA.
    :return: boolean whether the dependencies are satisfied.
    """
    if version == "latest":
        version = latest_commit(dataflow, user)
    if user:
        requirements_path = (
            USER_PATH.format(commit_ish=version, user=user, dataflow=dataflow) + "/requirements.txt"
        )
    else:
        requirements_path = (
            OFFICIAL_PATH.format(commit_ish=version, dataflow=dataflow) + "/requirements.txt"
        )

    if not os.path.exists(requirements_path):
        logger.info(f"requirements.txt not found at {requirements_path}")
        return True

    # Get list of currently installed packages
    with open(requirements_path, "r") as req_file:
        lines = req_file.readlines()

        for line in lines:
            line = line.strip()
            equals = line.find("=")
            less_than = line.find("<")
            greater_than = line.find(">")
            version_marker = min(equals, less_than, greater_than)
            if version_marker > 0:
                package_name, required_version = (
                    line[:version_marker],
                    line[version_marker + 1 :],
                )
            else:
                package_name = line
                required_version = None
            required_version  # here for now...
            try:
                installed_version = pkg_version(package_name)
                installed_version  # here for now..
            except PackageNotFoundError:
                logger.info(f"Package '{package_name}' is not installed.")
                return False

            # TODO: Check version if specified
            # if required_version and installed_version != required_version:
            #     logger.info(
            #         f"Package '{package_name}' version mismatch. Required: {required_version}, Installed: {installed_version}")
            #     return False

    logger.info("All requirements are satisfied.")
    return True


@_track_function_call
def list(version: str = "latest", user: str = None) -> list:
    """Lists dataflows locally downloaded based on commit_ish and user.

    :param version: the version to inspect. "latest" will resolve to the most recent commit, else pass \
         a commit SHA.
    :param user: the github name of the user.
    :return: list of tuples of (version, user, dataflow)
    """
    if version == "latest":
        version = "main"
    if not os.path.exists(DATAFLOW_FOLDER):
        # TODO better error message here.
        logger.info(f"Folder {DATAFLOW_FOLDER} does not exist.")
        return []

    dataflows = []

    for ci in os.listdir(DATAFLOW_FOLDER):
        if (
            (version and ci != version)
            or os.path.isfile(os.path.join(DATAFLOW_FOLDER, ci))
            or ci.startswith(".")
        ):
            continue
        commit_path = os.path.join(DATAFLOW_FOLDER, ci, "contrib", "hamilton", "contrib", "user")
        for usr in os.listdir(commit_path):
            if (
                (user and usr != user)
                or os.path.isfile(os.path.join(commit_path, usr))
                or usr.startswith(".")
            ):
                continue
            user_path = os.path.join(commit_path, usr)
            for df in os.listdir(user_path):
                if os.path.isfile(os.path.join(user_path, df)) or df.startswith("."):
                    continue
                dataflows.append((ci, usr, df))

    # for ci, usr, df in dataflows:
    #     logger.info(f"version: {ci}, user: {usr}, dataflow: {df}")

    return dataflows


@_track_function_call
def find(query: str, version: str = None, user: str = None):
    """Searches for locally downloaded dataflows based on a query string.

    :param query: key words to search for.
    :param version: the version to inspect. "latest" will resolve to the most recent commit, else pass \
         a commit SHA.
    :param user: the github name of the user.
    :return: list of tuples of (version, user, dataflow)
    """
    if not os.path.exists(DATAFLOW_FOLDER):
        logger.info(f"Folder {DATAFLOW_FOLDER} does not exist.")
        return []

    matches = set()

    for ci in os.listdir(DATAFLOW_FOLDER):
        if (
            (version and ci != version)
            or os.path.isfile(os.path.join(DATAFLOW_FOLDER, ci))
            or ci.startswith(".")
        ):
            continue
        commit_path = os.path.join(DATAFLOW_FOLDER, ci, "contrib", "hamilton", "contrib", "user")
        for usr in os.listdir(commit_path):
            if (
                (user and usr != user)
                or os.path.isfile(os.path.join(commit_path, usr))
                or usr.startswith(".")
            ):
                continue
            user_path = os.path.join(commit_path, usr)
            for df in os.listdir(user_path):
                for root, _, files in os.walk(user_path + "/" + df):
                    for file in files:
                        if not (
                            file.startswith(".") or file.endswith(".py") or file.endswith(".md")
                        ):
                            continue
                        file_path = os.path.join(root, file)
                        with open(file_path, "r") as f:
                            # dumb search -- just see if the string is in the file verbatim.
                            content = f.read()
                            if query in content:
                                matches.add((ci, usr, df))
    # TODO: incorporate tags
    return matches


@_track_function_call
def copy(
    dataflow: ModuleType,
    destination_path: str,
    overwrite: bool = False,
    renamed_module: str = None,
):
    """Copies a dataflow module to the passed in path.

    .. code-block:: python

        from hamilton import dataflows

        # dynamically pull and then copy
        NAME_OF_DATAFLOW = dataflow.import_module("NAME_OF_DATAFLOW", "NAME_OF_USER")
        dataflow.copy(NAME_OF_DATAFLOW, destination_path="PATH_TO_DIRECTORY")
        # copy from the installed library
        from hamilton.contrib.user.NAME_OF_USER import NAME_OF_DATAFLOW
        dataflow.copy(NAME_OF_DATAFLOW, destination_path="PATH_TO_DIRECTORY")

    :param dataflow: the module to copy.
    :param destination_path: the path to a directory to place the module in.
    :param overwrite: whether to overwrite the destination. Default is False and raise an error.
    :param renamed_module: whether to rename the copied module. Default is None and will use the original name.
    """
    # make sure the module exists and has a file to copy
    if not os.path.exists(dataflow.__file__):
        raise ValueError(f"Dataflow {dataflow.__name__} does not exist locally. Not copying.")

    if renamed_module:  # rename the module
        module_name = renamed_module
    else:
        # get the module name
        module_name = dataflow.__name__.split(".")[-1]

    # append the module name to the destination path
    destination_path = os.path.join(destination_path, module_name)
    # make the directories if they don't exist
    if not os.path.exists(destination_path):
        os.makedirs(destination_path, exist_ok=True)

    # Note: we'll need to change this if we change how module code is structured.
    file_path = os.path.join(destination_path, "__init__.py")
    # if the file_path exists and overwrite is False, raise an error
    if os.path.exists(file_path) and not overwrite:
        raise ValueError(
            f"Destination {file_path} already exists. Not copying. Use overwrite=True to overwrite."
        )
    # save the module code to the destination path
    shutil.copy(dataflow.__file__, file_path)
    logger.info(f"Successfully copied {dataflow.__name__} to {file_path}.")


@_track_function_call
def init():
    """Creates a template for someone to add a new flow for

    Don't want to bite off having to setup a fork,etc. This will
    just create a new directory with template files.
    """
    # TODO:
    pass


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO, stream=sys.stdout)
    _user = "zilto"
    _version = "latest"  # or a git commit hash
    _dataflow = "text_summarization"
    _module = import_module(_dataflow, _user, _version, overwrite=True)

    dr = driver.Driver({}, _module)
    # logger.info(dr.list_available_variables())
    dr.display_all_functions("./dag", {"format": "png", "view": False})

    logger.info("Listing dataflows ---")
    dfs = list(version=None, user="zilto")
    logger.info(dfs)
    logger.info("Listing chunks ---")
    chunks = find("chunk")
    logger.info(chunks)
    logger.info("Determining python dependencies ---")
    are_py_dependencies_satisfied(_user, _dataflow, _version)
    import pprint

    logger.info("Inspect module output ---")
    logger.info(pprint.pformat(inspect_module(_module)))
    logger.info("Resolve dataflow commit output ---")
    _version = latest_commit(_dataflow, _user)
    logger.info(_version)
    logger.info("Install dependencies output ---")
    logger.info(install_dependencies_string(_dataflow, _user, _version))
    logger.info("Inspect output ---")
    logger.info(inspect(_dataflow, _user, _version))

    from hamilton.contrib.user.zilto import text_summarization  # noqa:F401

    copy(
        text_summarization,
        destination_path="~/temp/",
        overwrite=True,
        renamed_module="text_summarization_copy",
    )
    copy(
        _module,
        destination_path="~/temp/",
        overwrite=True,
        renamed_module="text_summarization_copy2",
    )



from pathlib import Path

import dlt
import pandas as pd
import pyarrow as pa
import pytest
from dlt.destinations import filesystem

from hamilton.plugins.dlt_extensions import DltDestinationSaver, DltResourceLoader


def pandas_df():
    return pd.DataFrame({"a": [1, 2], "b": [1, 2]})


def iterable():
    return [{"a": 1, "b": 3}, {"a": 2, "b": 4}]


def pyarrow_table():
    col_a = pa.array([1, 2])
    col_b = pa.array([3, 4])
    return pa.Table.from_arrays([col_a, col_b], names=["a", "b"])


@pytest.mark.parametrize("data", [iterable(), pandas_df(), pyarrow_table()])
def test_dlt_destination_saver(data, tmp_path):
    save_pipe = dlt.pipeline(destination=filesystem(bucket_url=tmp_path.as_uri()))
    saver = DltDestinationSaver(pipeline=save_pipe, table_name="test_table")

    metadata = saver.save_data(data)

    assert len(metadata["dlt_metadata"]["load_packages"]) == 1
    assert metadata["dlt_metadata"]["load_packages"][0]["state"] == "loaded"
    assert Path(metadata["dlt_metadata"]["load_packages"][0]["jobs"][0]["file_path"]).exists()


def test_dlt_source_loader():
    resource = dlt.resource([{"a": 1, "b": 3}, {"a": 2, "b": 4}], name="mock_resource")
    loader = DltResourceLoader(resource=resource)

    loaded_data, metadata = loader.load_data(pd.DataFrame)

    assert len(loaded_data) == len([row for row in resource])
    assert "_dlt_load_id" in loaded_data.columns



import dataclasses
import pathlib
from typing import Any, Collection, Dict, Literal, Optional, Tuple, Type, Union

try:
    import mlflow
except ImportError:
    raise NotImplementedError("MLFlow is not installed.")

from hamilton import registry
from hamilton.io.data_adapters import DataLoader, DataSaver


@dataclasses.dataclass
class MLFlowModelSaver(DataSaver):
    """Save model to the MLFlow tracking server using `.log_model()`

    :param path: Run relative path to store model. Will constitute the model URI.
    :param register_as: If not None, register the model under the specified name.
    :param flavor: Library format to save the model (sklearn, xgboost, etc.). Automatically inferred if None.
    :param run_id: Log model to a specific run. Leave to `None` if using the `MLFlowTracker`
    :param kwargs: Arguments for `.log_model()`. Can be flavor-specific.
    """

    path: Union[str, pathlib.Path] = "model"
    register_as: Optional[str] = None
    flavor: Optional[str] = None
    run_id: Optional[str] = None
    kwargs: Dict[str, Any] = None

    def __post_init__(self):
        self.kwargs = self.kwargs if self.kwargs else {}

    @classmethod
    def name(cls) -> str:
        return "mlflow"

    @classmethod
    def applicable_types(cls) -> Collection[Type]:
        return [Any]

    def save_data(self, data) -> Dict[str, Any]:
        if self.flavor:
            flavor = self.flavor
        else:
            # infer the flavor from the base module of the data class
            # for example, extract `sklearn` from `sklearn.linear_model._base`
            flavor, _, _ = data.__module__.partition(".")

        # retrieve the `mlflow.FLAVOR` submodule to `.log_model()`
        try:
            flavor_module = getattr(mlflow, flavor)
        except ImportError:
            raise ImportError(f"Flavor {flavor} is unsupported by MLFlow")

        # handle `run_id` and active run conflicts
        if mlflow.active_run() and self.run_id:
            if mlflow.active_run().info.run_id != self.run_id:
                raise RuntimeError(
                    "The MLFlowModelSaver `run_id` doesn't match the active `run_id`\n",
                    "Set `run_id=None` to save to the active MLFlow run.",
                )

        # save to active run
        if mlflow.active_run():
            model_info = flavor_module.log_model(data, self.path, **self.kwargs)
        # create a run with `run_id` and save to it
        else:
            with mlflow.start_run(run_id=self.run_id):
                model_info = flavor_module.log_model(data, self.path, **self.kwargs)

        # create metadata from ModelInfo object
        metadata = {k.strip("_"): v for k, v in model_info.__dict__.items()}

        if self.register_as:
            model_version = mlflow.register_model(
                model_uri=metadata["model_uri"], name=self.register_as
            )
            # update metadata with the registered ModelVersion
            # there's a contract between this key and the MLFlowTracker's
            # post_node_execute() reads this metadata
            metadata["registered_model"] = {
                k.strip("_"): v for k, v in model_version.__dict__.items()
            }

        return metadata


@dataclasses.dataclass
class MLFlowModelLoader(DataLoader):
    """Load model from the MLFlow tracking server or model registry using .load_model()
    You can pass a model URI or the necessary metadata to retrieve the model

    :param model_uri: Model location starting as `runs:/` for tracking or `models:/` for registry
    :param mode: `tracking` or registry`. tracking needs `run_id` and `path`. registry needs `model_name` and `version` or `version_alias`.
    :param run_id: Run id of the model on the tracking server
    :param path: Run relative path where the model is stored
    :param model_name: Name of the registered model (equivalent to `register_as` in model saver)
    :param version: Version of the registered model. Can pass as string `v1` or integer `1`
    :param version_alias: Version alias of the registered model. Specify either this or `version`
    :param flavor: Library format to load the model (sklearn, xgboost, etc.). Automatically inferred if None.
    :param kwargs: Arguments for `.load_model()`. Can be flavor-specific.
    """

    model_uri: Optional[str] = None
    mode: Literal["tracking", "registry"] = "tracking"
    run_id: Optional[str] = None
    path: Union[str, pathlib.Path] = "model"
    model_name: Optional[str] = None
    version: Optional[Union[str, int]] = None
    version_alias: Optional[str] = None
    flavor: Optional[str] = None
    kwargs: Dict[str, Any] = None

    # __post_init__ is required to set kwargs as empty dict because
    # can't set: kwargs: Dict[str, Any] = dataclasses.field(default_factory=dict)
    # otherwise raises `InvalidDecoratorException` because materializer factory check
    # for all params being set and `kwargs` would be unset until instantiation.
    def __post_init__(self):
        self.kwargs = self.kwargs if self.kwargs else {}

        if self.model_uri:
            return

        if self.mode == "tracking":
            if (not self.run_id) or (not self.path):
                raise ValueError("Using `mode='tracking'` requires passing `run_id` and `path`")

            self.model_uri = f"runs:/{self.run_id}/{self.path}"

        elif self.mode == "registry":
            if not self.model_name:
                raise ValueError("Using `mode='registry` requires passing `model_name`")

            if not (bool(self.version) ^ bool(self.version_alias)):
                raise ValueError(
                    "If using `mode='registry'` requires passing `version` OR `version_alias"
                )

            if self.version:
                self.model_uri = f"models:/{self.model_name}/{self.version}"
            elif self.version_alias:
                self.model_uri = f"models:/{self.model_name}@{self.version_alias}"

    @classmethod
    def name(cls) -> str:
        return "mlflow"

    @classmethod
    def applicable_types(cls) -> Collection[Type]:
        return [Any]

    def load_data(self, type_: Type) -> Tuple[Any, Dict[str, Any]]:
        model_info = mlflow.models.model.get_model_info(self.model_uri)
        metadata = {k.strip("_"): v for k, v in model_info.__dict__.items()}

        flavor = self.flavor
        # if flavor not explicitly passed, retrieve flavor from the ModelInfo
        if flavor is None:
            # prioritize the library specific flavor. Default to `pyfunc` if none available.
            try:
                flavor = next(f for f in metadata["flavors"].keys() if f != "python_function")
            except StopIteration:
                flavor = "pyfunc"

        # retrieve the `mlflow.FLAVOR` submodule to `.log_model()`
        try:
            flavor_module = getattr(mlflow, flavor)
        except ImportError:
            raise ImportError(f"Flavor {flavor} is unsupported by MLFlow")

        model = flavor_module.load_model(model_uri=self.model_uri)
        return model, metadata


def register_data_loaders():
    """Function to register the data loaders for this extension."""
    for loader in [
        MLFlowModelSaver,
        MLFlowModelLoader,
    ]:
        registry.register_adapter(loader)


register_data_loaders()

COLUMN_FRIENDLY_DF_TYPE = False



from typing import Any, Dict, List, Type, Union

import numpy as np
import pandas as pd

from hamilton import base

try:
    import vaex
except ImportError:
    raise NotImplementedError("Vaex is not installed.")


class VaexDataFrameResult(base.ResultMixin):
    """A ResultBuilder that produces a Vaex dataframe.

    Use this when you want to create a Vaex dataframe from the outputs.
    Caveat: you need to ensure that the length
    of the outputs is the same (except scalars), otherwise you will get an error;
    mixed outputs aren't that well handled.

    To use:

    .. code-block:: python

        from hamilton import base, driver
        from hamilton.plugins import h_vaex
        vaex_builder = h_vaex.VaexDataFrameResult()
        adapter = base.SimplePythonGraphAdapter(vaex_builder)
        dr =  driver.Driver(config, *modules, adapter=adapter)
        df = dr.execute([...], inputs=...)  # returns vaex dataframe

    Note: this is just a first attempt at something for Vaex.
    Think it should handle more? Come chat/open a PR!
    """

    def build_result(
        self,
        **outputs: Dict[str, Union[vaex.expression.Expression, vaex.dataframe.DataFrame, Any]],
    ):
        """This is the method that Hamilton will call to build the final result.
        It will pass in the results of the requested outputs that
        you passed in to the execute() method.

        :param outputs: The results of the requested outputs.
        :return: a Vaex DataFrame.
        """

        # We split all outputs into DataFrames, arrays and scalars
        dfs: List[vaex.dataframe.DataFrame] = []  # Vaex DataFrames from outputs
        arrays: Dict[str, np.ndarray] = dict()  # arrays from outputs
        scalars: Dict[str, Any] = dict()  # scalars from outputs

        for name, value in outputs.items():
            if isinstance(value, vaex.dataframe.DataFrame):
                dfs.append(value)
            elif isinstance(value, vaex.expression.Expression):
                nparray = value.to_numpy()
                if nparray.ndim == 0:  # value is scalar
                    scalars[name] = nparray.item()
                elif nparray.shape == (1,):  # value is scalar
                    scalars[name] = nparray[0]
                else:  # value is array
                    arrays[name] = nparray
            elif isinstance(value, np.ndarray):
                if value.ndim == 0:  # value is scalar
                    scalars[name] = value.item()
                elif value.shape == (1,):  # value is scalar
                    scalars[name] = value[0]
                else:  # value is array
                    arrays[name] = value
            elif pd.api.types.is_scalar(value):  # value is scalar
                scalars[name] = value
            else:
                value_type = str(type(value))
                message = f"VaexDataFrameResult doesn't support {value_type}"
                raise NotImplementedError(message)

        df = None

        if arrays:

            # Check if all arrays have correct and identical shapes.
            first_expression_shape = next(arrays.values().__iter__()).shape
            if len(first_expression_shape) > 1:
                raise NotImplementedError(
                    "VaexDataFrameResult supports only one-dimensional Expression results"
                )
            for name, a in arrays.items():
                if a.shape != first_expression_shape:
                    raise NotImplementedError(
                        "VaexDataFrameResult supports Expression results with same dimension only"
                    )

            # All scalars become arrays with the same shape as other arrays.
            for name, scalar in scalars.items():
                arrays[name] = np.full(first_expression_shape, scalar)

            df = vaex.from_arrays(**arrays)

        elif scalars:

            # There are not arrays in outputs,
            # so we construct Vaex DataFrame with one row consisting of scalars.
            df = vaex.from_arrays(**{name: np.array([value]) for name, value in scalars.items()})

        if df:
            dfs.append(df)

        return vaex.concat(dfs)

    def output_type(self) -> Type:
        return vaex.dataframe.DataFrame



import inspect
from typing import Any, Dict, List, Optional, Tuple, Type

from kedro.pipeline.node import Node as KNode
from kedro.pipeline.pipeline import Pipeline as KPipeline

from hamilton import driver, graph
from hamilton.function_modifiers.expanders import extract_fields
from hamilton.lifecycle import base as lifecycle_base
from hamilton.node import Node as HNode


def expand_k_node(base_node: HNode, outputs: List[str]) -> List[HNode]:
    """Manually apply `@extract_fields()` on a Hamilton node.Node for a Kedro
    node that specifies >1 `outputs`.

    The number of nodes == len(outputs) + 1 because it includes the `base_node`
    """

    def _convert_output_from_tuple_to_dict(node_result: Any, node_kwargs: Dict[str, Any]):
        return {out: v for out, v in zip(outputs, node_result)}

    # NOTE isinstance(Any, type) is False for Python < 3.11
    extractor = extract_fields(fields={out: Any for out in outputs})
    func = base_node.originating_functions[0]
    if issubclass(func.__annotations__["return"], Tuple):
        base_node = base_node.transform_output(_convert_output_from_tuple_to_dict, Dict)
        func.__annotations__["return"] = Dict

    extractor.validate(func)
    return list(extractor.transform_node(base_node, {}, func))


def k_node_to_h_nodes(node: KNode) -> List[HNode]:
    """Convert a Kedro node to a list of Hamilton nodes.
    If the Kedro node specifies 1 output, generate 1 Hamilton node.
    If it generate >1 output, generate len(outputs) + 1 to include the base node + extracted fields.
    """
    # determine if more than one output
    node_names = []
    if isinstance(node.outputs, list):
        node_names.extend(node.outputs)
    elif isinstance(node.outputs, dict):
        node_names.extend(node.outputs.values())

    # determine the base node name
    if len(node_names) == 1:
        base_node_name = node_names[0]
    elif isinstance(node.outputs, str):
        base_node_name = node.outputs
    else:
        base_node_name = node.func.__name__

    func_sig = inspect.signature(node.func)
    params = func_sig.parameters.values()
    output_type = func_sig.return_annotation
    if output_type is None:
        # manually creating `hamilton.node.Node` doesn't accept `typ=None`
        output_type = Type[None]  # NoneType is introduced in Python 3.10

    base_node = HNode(
        name=base_node_name,
        typ=output_type,
        doc_string=getattr(node.func, "__doc__", ""),
        callabl=node.func,
        originating_functions=(node.func,),
    )

    # if Kedro node defines multiple outputs, use `@extract_fields()`
    if len(node_names) > 1:
        h_nodes = expand_k_node(base_node, node_names)
    else:
        h_nodes = [base_node]

    # remap the function parameters to the node `inputs` and clean Kedro `parameters` name
    new_params = {}
    for param, k_input in zip(params, node.inputs):
        if k_input.startswith("params:"):
            k_input = k_input.partition("params:")[-1]

        new_params[param.name] = k_input

    h_nodes = [n.reassign_inputs(input_names=new_params) for n in h_nodes]

    return h_nodes


def kedro_pipeline_to_driver(
    *pipelines: KPipeline,
    builder: Optional[driver.Builder] = None,
) -> driver.Driver:
    """Convert one or mode Kedro `Pipeline` to a Hamilton `Driver`.
    Pass a Hamilton `Builder` to include lifecycle adapters in your `Driver`.

    :param pipelines: one or more Kedro `Pipeline` objects
    :param builder: a Hamilton `Builder` to use when building the `Driver`
    :return: the Hamilton `Driver` built from Kedro `Pipeline` objects.

    .. code-block: python

        from hamilton import driver
        from hamilton.plugins import h_kedro

        builder = driver.Builder().with_adapters(tracker)

        dr = h_kedro.kedro_pipeline_to_driver(
            data_science.create_pipeline(),  # Kedro Pipeline
            data_processing.create_pipeline(),  # Kedro Pipeline
            builder=builder
        )
    """
    # generate nodes
    h_nodes = []
    for pipe in pipelines:
        for node in pipe.nodes:
            h_nodes.extend(k_node_to_h_nodes(node))

    # resolve dependencies
    h_nodes = graph.update_dependencies(
        {n.name: n for n in h_nodes},
        lifecycle_base.LifecycleAdapterSet(),
    )

    builder = builder if builder else driver.Builder()
    dr = builder.build()
    # inject function graph in Driver
    dr.graph = graph.FunctionGraph(
        h_nodes, config={}, adapter=lifecycle_base.LifecycleAdapterSet(*builder.adapters)
    )
    # reapply lifecycle hooks
    if dr.adapter.does_hook("post_graph_construct", is_async=False):
        dr.adapter.call_all_lifecycle_hooks_sync(
            "post_graph_construct", graph=dr.graph, modules=dr.graph_modules, config={}
        )
    return dr



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import warnings

"""The setup script."""

from setuptools import find_packages, setup

# don't fail if there are problems with the readme (happens within circleci)
try:
    with open("README.md") as readme_file:
        readme = readme_file.read()
except Exception:
    warnings.warn("README.md not found")
    readme = None

REQUIREMENTS_FILES = ["requirements.txt"]


def get_version():
    version_dict = {}
    with open("hamilton/version.py") as f:
        exec(f.read(), version_dict)
    return ".".join(map(str, version_dict["VERSION"]))


VERSION = get_version()


def load_requirements():
    requirements = set()
    for requirement_file in REQUIREMENTS_FILES:
        with open(requirement_file) as f:
            requirements.update(line.strip() for line in f)
    return list(requirements)


setup(
    name="sf-hamilton",  # there's already a hamilton in pypi
    version=VERSION,
    description="Hamilton, the micro-framework for creating dataframes.",
    long_description=readme,
    long_description_content_type="text/markdown",
    author="Stefan Krawczyk, Elijah ben Izzy",
    author_email="stefan@dagworks.io,elijah@dagworks.io",
    url="https://github.com/dagworks-inc/hamilton",
    packages=find_packages(exclude=["tests"], include=["hamilton", "hamilton.*"]),
    include_package_data=True,
    install_requires=load_requirements(),
    zip_safe=False,
    keywords="hamilton",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
    # Note that this feature requires pep8 >= v9 and a version of setup tools greater than the
    # default version installed with virtualenv. Make sure to update your tools!
    python_requires=">=3.6, <4",
    # adding this to slim the package down, since these dependencies are only used in certain contexts.
    extras_require={
        "visualization": ["graphviz", "networkx"],
        "dask": ["dask[complete]"],  # commonly you'll want everything.
        "dask-array": ["dask[array]"],
        "dask-core": ["dask-core"],
        "dask-dataframe": ["dask[dataframe]"],
        "dask-diagnostics": ["dask[diagnostics]"],
        "dask-distributed": ["dask[distributed]"],
        "ray": ["ray>=2.0.0", "pyarrow"],
        "pyspark": [
            # we have to run these dependencies cause Spark does not check to ensure the right target was called
            "pyspark[pandas_on_spark,sql]",
            # This is problematic, see https://stackoverflow.com/questions/76072664/convert-pyspark-dataframe-to-pandas-dataframe-fails-on-timestamp-column
            "pandas<2.0",
        ],  # I'm sure they'll add support soon,
        # but for now its not compatible
        "pandera": ["pandera"],
        "slack": ["slack-sdk"],
        "tqdm": ["tqdm"],
        "datadog": ["ddtrace"],
        "vaex": [
            "pydantic<2.0",  # because of https://github.com/vaexio/vaex/issues/2384
            "vaex",
        ],
        "experiments": [
            "fastapi",
            "fastui",
            "uvicorn",
        ],
        "diskcache": ["diskcache"],
        "cli": ["typer"],
        "sdk": ["sf-hamilton-sdk"],
        "ui": ["sf-hamilton-ui"],
    },
    entry_points={
        "console_scripts": [
            "h_experiments = hamilton.plugins.h_experiments.__main__:main",
            "hamilton = hamilton.cli.__main__:cli",
            "hamilton-admin-build-ui = hamilton.admin:build_ui",
            "hamilton-admin-build-and-publish = hamilton.admin:build_and_publish",
        ]
    },
    # Relevant project URLs
    project_urls={  # Optional
        "Bug Reports": "https://github.com/dagworks-inc/hamilton/issues",
        "Source": "https://github.com/dagworks-inc/hamilton",
    },
)



import dataclasses
import functools
import inspect
import typing
from typing import Any, Dict, List, Optional, Protocol, Set, Type, Union

from hamilton import base, common, graph, lifecycle, node
from hamilton.function_modifiers.adapters import LoadFromDecorator, SaveToDecorator
from hamilton.function_modifiers.dependencies import SingleDependency, value
from hamilton.graph import FunctionGraph, update_dependencies
from hamilton.io.data_adapters import DataLoader, DataSaver
from hamilton.registry import LOADER_REGISTRY, SAVER_REGISTRY


class materialization_meta__(type):
    """Metaclass for the to. materializer. This is specifically to allow class access method.
    This only exists to add more helpful error messages. We dynamically assign the attributes
    below (see `_set_materializer_attrs`), which helps with auto-complete.
    """

    def __new__(cls, name, bases, clsdict):
        """Boiler plate for a metaclass -- this just instantiates it as a type. It also sets the
        annotations to be available later.
        """
        clsobj = super().__new__(cls, name, bases, clsdict)
        clsobj.__annotations__ = {}
        return clsobj

    def __getattr__(cls, item: str) -> "_MaterializerFactoryProtocol":
        """This *just* exists to provide a more helpful error message. If you try to access
        a property that doesn't exist, we'll raise an error that tells you what properties
        are available/where to learn more."""

        try:
            return super().__getattribute__(item)
        except AttributeError as e:
            if item in SAVER_REGISTRY:
                # In this case we want to dynamically access it after registering
                # This is so that we can register post-importing
                # Note that this will register all new ones in bulk --
                # which will (in most cases) be a one-time-cost
                _set_materializer_attrs()
                return super().__getattribute__(item)
            raise AttributeError(
                f"No data materializer named: {item}. "
                f"Available materializers are: {list(SAVER_REGISTRY.keys())}. "
                "If you've gotten to this point, you either (1) Have created the custom materializer and not "
                "registered it (in which case, call registry.register_adapter...) "
                "(2) spelled the materializer name wrong, or (e) are trying to use a materialiazer that does"
                "not exist (yet). For a list of available materialiazers, see  "
                "https://hamilton.readthedocs.io/reference/io/available-data-adapters/#data"
                "-loaders "
            ) from e


class extractor_meta__(type):
    """Metaclass for the from_. extractor pattern. This is specifically to allow class access method.
    This only exists to add more helpful error messages. We dynamically assign the attributes
    below (see `_set_extractor_attrs`), which helps with auto-complete.

    TODO -- reduce shared code between this and the loader above.
    """

    def __new__(cls, name, bases, clsdict):
        """Boiler plate for a metaclass -- this just instantiates it as a type. It also sets the
        annotations to be available later.
        """
        clsobj = super().__new__(cls, name, bases, clsdict)
        clsobj.__annotations__ = {}
        return clsobj

    def __getattr__(cls, item: str) -> "_ExtractorFactoryProtocol":
        """This *just* exists to provide a more helpful error message. If you try to access
        a property that doesn't exist, we'll raise an error that tells you what properties
        are available/where to learn more."""
        try:
            return super().__getattribute__(item)
        except AttributeError as e:
            if item in LOADER_REGISTRY:
                # See note on data savers/__getattr__ above
                _set_materializer_attrs()
                return super().__getattribute__(item)
            raise AttributeError(
                f"No data loader named: {item}. "
                f"Available loaders are: {LOADER_REGISTRY.keys()}. "
                "If you've gotten to this point, you either (1) spelled the "
                "loader name wrong, (2) are trying to use a loader that does"
                "not exist (yet). For a list of available loaders, see  "
                "https://hamilton.readthedocs.io/reference/io/available-data-adapters/#data"
                "-loaders "
            ) from e


def process_kwargs(
    data_saver_kwargs: Dict[str, Union[Any, SingleDependency]]
) -> Dict[str, SingleDependency]:
    """Processes raw strings from the user, converting them into dependency specs.
    This goes according to the following rules.

    1. If it is a `SingleDependency`, we leave it alone
    2. If it is anything else, we treat it as a literal, and convert it to a value(...)

    This is so that everything is in the shape of source/value, and we can uniformally handle it later.

    :param data_saver_kwargs: Kwargs passed in from the user
    :return: Processed kwargs
    """
    processed_kwargs = {}
    for kwarg, kwarg_val in data_saver_kwargs.items():
        if not isinstance(kwarg_val, SingleDependency):
            processed_kwargs[kwarg] = value(kwarg_val)
        else:
            processed_kwargs[kwarg] = kwarg_val
    return processed_kwargs


class ExtractorFactory:
    def __init__(
        self,
        target: str,
        loaders: List[Type[DataLoader]],
        **data_loader_kwargs: Union[Any, SingleDependency],
    ):
        """Instantiates an ExtractorFactory. Note this is not a public API -- this is
        internally what gets called (through a factory method) to create it. Called using `from_`,
        E.G. `from_.csv`.

        :param target: Parameter, into which we're loading the data
        :param loaders: A list of data loaders that are viable candidates, given the key after `from_`
        :param data_loader_kwargs: Keyword arguments for the data loaders.
        """
        self.target = target
        self.loaders = loaders
        self.data_loader_kwargs = process_kwargs(data_loader_kwargs)

    def generate_nodes(self, fn_graph: graph.FunctionGraph) -> List[node.Node]:
        """Resolves the extractor, returning the set of nodes that should get
        added to the function graph. Note that this is an upsert operation --
        these nodes can replace existing nodes.

        :param fn_graph: Function graph
        :return: List of nodes to add/upsert to the graph
        """

        decorator = LoadFromDecorator(self.loaders, self.target, **self.data_loader_kwargs)
        # TODO -- add some nodes to the graph
        node_with_target = fn_graph.nodes.get(self.target)
        if node_with_target is None:
            raise ValueError(
                f"Could not find node with name: {self.target} in function "
                f"graph. Available nodes: {list(fn_graph.nodes.keys()) + [...] if len(fn_graph.nodes) > 10 else []}"
            )
        return decorator.get_loader_nodes(self.target, node_with_target.type, namespace=None)


class MaterializerFactory:
    """Basic factory for creating materializers. Note that this should only ever be instantiated
    through `to.<name>`, which conducts polymorphic lookup to find the appropriate materializer.
    """

    def __init__(
        self,
        id: str,
        savers: List[Type[DataSaver]],
        result_builder: Optional[base.ResultMixin],
        dependencies: List[Union[str, Any]],
        **data_saver_kwargs: Any,
    ):
        """Creates a materializer factory.

        :param name: Name of the node that this will represent in the DAG.
        :param savers: Potential data savers to use (these will be filtered down from the ones
        registered to <output_format> when `to.<output_format>` is called.)
        :param result_builder: ResultBuilder that joins the result of the dependencies together.
        :param dependencies: Nodes, the results of which on which this depends
        :param data_saver_kwargs: kwargs to be passed to the data_savers. Either literal values,
        or `source`/`value`.
        """

        self.id = id
        self.savers = savers
        self.result_builder = result_builder
        self.dependencies = dependencies
        self.data_saver_kwargs = process_kwargs(data_saver_kwargs)

    def sanitize_dependencies(self, module_set: Set[str]) -> "MaterializerFactory":
        """Sanitizes the dependencies to ensure they're strings.

        This replaces the internal value for self.dependencies and returns a new object.
        We return a new object to not modify the one passed in.

        :param module_set: modules that "functions" could come from if that's passed in.
        :return: new object with sanitized_dependencies.
        """
        final_vars = common.convert_output_values(self.dependencies, module_set)
        return MaterializerFactory(
            self.id,
            self.savers,
            self.result_builder,
            final_vars,
            **self.data_saver_kwargs,
        )

    def _resolve_dependencies(self, fn_graph: graph.FunctionGraph) -> List[node.Node]:
        out = []
        missing_nodes = []
        for name in self.dependencies:
            if name in fn_graph.nodes:
                out.append(fn_graph.nodes[name])
            else:
                missing_nodes.append(name)
        if missing_nodes:
            raise ValueError(
                f"Materializer {self.id} has dependencies that are not in the graph: {missing_nodes}."
            )
        return [fn_graph.nodes[name] for name in self.dependencies]

    def generate_nodes(self, fn_graph: graph.FunctionGraph) -> List[node.Node]:
        """Generates additional nodes from a materializer, returning the set of nodes that should get
        appended to the function graph. This does two things:

        1. Adds a node that handles result-building
        2. Adds a node that handles data-saving, reusing the data saver functionality.

        :param graph: Function Graph to which we are adding the materializer
        :return: List of nodes to add to the graph
        """
        node_dependencies = self._resolve_dependencies(fn_graph)

        def join_function(**kwargs):
            return self.result_builder.build_result(**kwargs)

        out = []
        if self.result_builder is None:
            if len(node_dependencies) != 1:
                raise ValueError(
                    "Must specify result builder via combine= key word argument if the materializer has more than "
                    "one dependency it is materializing. Otherwise we have no way to join them and know what to pass! "
                    f"See materializer {self.id}."
                )
            save_dep = node_dependencies[0]
        else:
            join_node = node.Node(
                name=f"{self.id}_build_result",
                typ=self.result_builder.output_type(),
                doc_string=f"Builds the result for {self.id} materializer",
                callabl=join_function,
                input_types={dep.name: dep.type for dep in node_dependencies},
                originating_functions=(
                    None if self.result_builder is None else [self.result_builder.build_result]
                ),
            )
            out.append(join_node)
            save_dep = join_node

        out.append(
            # We can reuse the functionality in the save_to decorator
            SaveToDecorator(self.savers, self.id, **self.data_saver_kwargs).create_saver_node(
                save_dep, {}, save_dep.callable
            )
        )
        return out


@typing.runtime_checkable
class _MaterializerFactoryProtocol(Protocol):
    """Typing for the create_materializer_factory function"""

    def __call__(
        self,
        id: str,
        dependencies: List[str],
        combine: lifecycle.ResultBuilder = None,
        **kwargs: Union[str, SingleDependency],
    ) -> MaterializerFactory:
        pass


@typing.runtime_checkable
class _ExtractorFactoryProtocol(Protocol):
    def __call__(self, target: str, **kwargs: Union[str, SingleDependency]) -> ExtractorFactory:
        pass


def partial_materializer(data_savers: List[Type[DataSaver]]) -> _MaterializerFactoryProtocol:
    """Creates a partial materializer, with the specified data savers."""

    def create_materializer_factory(
        id: str,
        dependencies: List[str],
        combine: base.ResultMixin = None,
        **kwargs: typing.Any,
    ) -> MaterializerFactory:
        return MaterializerFactory(
            id=id,
            savers=data_savers,
            result_builder=combine,
            dependencies=dependencies,
            **kwargs,
        )

    return create_materializer_factory


def partial_extractor(
    data_loaders: List[Type[DataLoader]],
) -> _ExtractorFactoryProtocol:
    """Creates a partial materializer, with the specified data savers."""

    def create_extractor_factory(
        target: str,
        **kwargs: typing.Any,
    ) -> ExtractorFactory:
        return ExtractorFactory(
            target=target,
            loaders=data_loaders,
            **kwargs,
        )

    return create_extractor_factory


class Materialize(metaclass=materialization_meta__):
    """Materialize class to facilitate easy reference. Note that you should never need to refer
    to this directly. Rather, this should be referred as `to` in hamilton.io."""


class to(Materialize):
    """This is the entry point for Materialization. Note that this is coupled
    with the driver's materialize function -- properties are dynamically assigned
    based on data savers that have been registered, allowing you to call `to.csv`/`to.json`.
    For full documentation, see the documentation for the `materialize` function in the hamilton
    driver."""

    pass


class Extract(metaclass=extractor_meta__):
    """Extract class to facilitate easy reference. Note that you should never need to refer
    to this directly. Rather, this should be referred as `from_` in hamilton.io."""


class from_(Extract):
    """This is the entry point for Extraction. Note that this is coupled
    with the driver's materialiez function -- properties are dynamically assigned
    based on data loaders that have been registered, allowing you to call `from_.csv`/`from_.json`.
    For full documentation, see the documentation for the `materialize` function in the hamilton
    driver.
    """


def _set_materializer_attrs():
    """Sets materialization attributes for easy reference. This sets it to the available keys.
    This is so one can get auto-complete"""

    def with_modified_signature(
        fn: Type[_MaterializerFactoryProtocol],
        dataclasses_union: List[Type[dataclasses.dataclass]],
        key: str,
    ):
        """Modifies the signature to include the parameters from *all* dataclasses.
        Note this just replaces **kwargs with the union of the parameters. Its not
        strictly correct, as (a) its a superset of the available ones and (b) it doesn't
        include the source/value parameters. However, this can help the development experience
        on jupyter notebooks, and is a good enough approximation for now.


        :param fn: Function to modify -- will change the signature.
        :param dataclasses_union: All dataclasses to union.
        :param key: Key to use for the materializer.
        :return: The function without **kwargs and with the union of the parameters.
        """
        original_signature = inspect.signature(fn)
        original_parameters = list(original_signature.parameters.values())

        new_parameters = []
        seen = set()
        for dataclass in dataclasses_union:
            for field in dataclasses.fields(dataclass):
                if field.name not in seen:
                    new_parameters.append(
                        inspect.Parameter(
                            field.name,
                            inspect.Parameter.KEYWORD_ONLY,
                            annotation=field.type,
                            default=None,
                        )
                    )
                seen.add(field.name)

        # Combining old and new parameters
        # Checking for position of **kwargs and insert new params before
        for idx, param in enumerate(original_parameters):
            if param.kind == inspect.Parameter.VAR_KEYWORD:
                break
        else:
            idx = len(original_parameters)

        # Insert new parameters while respecting the order
        combined_parameters = original_parameters[:idx] + new_parameters + original_parameters[idx:]
        combined_parameters = [param for param in combined_parameters if param.name != "kwargs"]

        # Creating a new signature with combined parameters
        new_signature = original_signature.replace(parameters=combined_parameters)

        # Creating a new function with the new signature
        @functools.wraps(fn)
        def wrapper(*args, **kwargs):
            bound_arguments = new_signature.bind(*args, **kwargs)
            return fn(*bound_arguments.args, **bound_arguments.kwargs)

        # Assign the new signature to the wrapper function
        wrapper.__signature__ = new_signature
        wrapper.__doc__ = f"""
        Materializes data to {key} format. Note that the parameters are a superset of possible parameters -- this might depend on
        the actual type of the data passed in. For more information, see: https://hamilton.dagworks.io/en/latest/reference/io/available-data-adapters/#data-loaders.
        You can also pass `source` and `value` in as kwargs.
        """
        return wrapper

    # Go through savers and loaders and add them to the class
    # This way we can access with from_.xyz/to.xyz
    for registry, cls_target, adapter_type, partial_factory in [
        (SAVER_REGISTRY, Materialize, DataSaver, partial_materializer),
        (LOADER_REGISTRY, Extract, DataLoader, partial_extractor),
    ]:
        for key, potential_loaders in registry.items():
            loaders = [loader for loader in potential_loaders if issubclass(loader, adapter_type)]
            if len(loaders) > 0:
                partial = partial_factory(potential_loaders)
                partial_with_signature = with_modified_signature(partial, potential_loaders, key)
                setattr(cls_target, key, partial_with_signature)
                cls_target.__annotations__[key] = type(partial_with_signature)


_set_materializer_attrs()


def modify_graph(
    fn_graph: FunctionGraph,
    materializer_factories: List[MaterializerFactory],
    extractor_factories: List[ExtractorFactory],
) -> FunctionGraph:
    """Modifies the function graph, adding in the specified materialization/loader nodes.

    Note that this is not simple -- adding materializers is different from adding loaders:
    1. Materializers can simply be appended to the beginning of the graph
    2. loaders (currently) function as "injected nodes", meaning they behave as overrides.
    We don't actually *feed* them in as overrides (as that would have to be a 2-pass process).
    Rather, we add them to the graph, then prune the paths that are upstream of *just them*,
    that can't possibly be executed

    :param graph: Graph to modify.
    :param materializer_factories: Materializer factories (created by to.xyz) to add to the graph
    :param extractor_factories: Loader factories (created by from_.xyz) to add to the graph
    :return: A new graph with the materializers.
    """
    materializer_nodes = []
    for materializer in materializer_factories:
        materializer_nodes.extend(materializer.generate_nodes(fn_graph))
    fn_graph = fn_graph.with_nodes({node_.name: node_ for node_ in materializer_nodes})
    loader_nodes = []
    # We want to treat this as an override
    # For now what we'll do is:
    # 1. Replace the nodes we're replacing
    # 2. Update dependencies
    # This will leave some dangling nodes, but we can deal with those later...
    for loader in extractor_factories:
        loader_nodes.extend(loader.generate_nodes(fn_graph))
    graph_nodes = fn_graph.nodes.copy()
    for loader_node in loader_nodes:
        graph_nodes[loader_node.name] = loader_node
    # Simpler to just create a new one
    # This leaks some details slightly -- E.G. how the dependencies work
    # We should probably add this as part of the functiong raph constructor?
    # For now this will be OK
    fn_graph = graph.FunctionGraph(
        nodes=update_dependencies(graph_nodes, fn_graph.adapter),
        config=fn_graph.config,
        adapter=fn_graph.adapter,
    )
    # TODO -- prune the nodes that are upstream *only* of the old loader nodes, and not
    return fn_graph
